from wbcore import viewsets

from wbportfolio.models import PortfolioCashTarget
from wbportfolio.serializers import PortfolioCashTargetModelSerializer


class PortfolioCashTargetModelViewSet(viewsets.ModelViewSet):
    queryset = PortfolioCashTarget.objects.all()
    serializer_class = PortfolioCashTargetModelSerializer



import factory

from .models import AssetAllocation, AssetAllocationType, GeographicFocus


class AssetAllocationTypeFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("word")
    default_max_investment = factory.Faker("pydecimal", min_value=0.0, max_value=1, right_digits=4)

    class Meta:
        model = AssetAllocationType


class AssetAllocationFactory(factory.django.DjangoModelFactory):
    company = factory.SubFactory("wbcore.contrib.directory.factories.CompanyFactory")
    asset_type = factory.SubFactory(AssetAllocationTypeFactory)
    percent = factory.Faker("pydecimal", min_value=0.0, max_value=1, right_digits=4)
    max_investment = factory.Faker("pydecimal", min_value=0.0, max_value=1, right_digits=4)
    comment = factory.Faker("sentence")

    class Meta:
        model = AssetAllocation


class GeographicFocusFactory(factory.django.DjangoModelFactory):
    company = factory.SubFactory("wbcore.contrib.directory.factories.CompanyFactory")
    country = factory.SubFactory("wbcore.contrib.geography.factories.CountryFactory")
    percent = factory.Faker("pydecimal", min_value=0.0, max_value=1, right_digits=4)
    comment = factory.Faker("sentence")

    class Meta:
        model = GeographicFocus



from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.enums import ButtonDefaultColor
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class AccountReconciliationButtonViewConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        return {
            bt.ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.SUCCESS,
                identifiers=("wbportfolio:accountreconciliation",),
                icon=WBIcon.APPROVE.icon,
                key="agree_customer",
                label="Agree",
                action_label="Agreeing",
                description_fields="<p>After agreening to this holdings, you cannot change this reconcilation anymore!</p>",
            ),
            bt.ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.ERROR,
                identifiers=("wbportfolio:accountreconciliation",),
                icon=WBIcon.DENY.icon,
                key="disagree_customer",
                label="Communicate Differences",
                action_label="Communicating Differences",
                description_fields="<p>After disagreeing with this holdings, someone will reach out to you do discuss the differences.</p>",
            ),
            bt.ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.ERROR,
                identifiers=("wbportfolio:accountreconciliationline", "wbportfolio:accountreconciliation"),
                icon=WBIcon.REFRESH.icon,
                key="recalculate",
                label="Recalculate",
                action_label="Recalculating",
                description_fields="<p>After confirming, the Reconciliation will recompute 'Our Calculations'</p>",
            ),
            bt.ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbportfolio:accountreconciliationline",),
                icon=WBIcon.MAIL.icon,
                key="notify",
                label="Notify Reconciliation Managers",
                action_label="Notifying",
                description_fields="<p>After confirming the reconciliation managers of this account reconciliation will be notified.</p>",
            ),
            bt.WidgetButton(key="claims", label="Show Subscriptions/Redemptions", icon=WBIcon.TRADE.icon, weight=200),
            bt.WidgetButton(
                key="add-claim", label="Add Subscription/Redemption", icon=WBIcon.ADD.icon, new_mode=True, weight=300
            ),
        }


class AccountReconciliationLineButtonViewConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(key="product-claims", label="Show Subscriptions/Redemptions", icon=WBIcon.TRADE.icon),
            bt.WidgetButton(
                key="add-product-claim", label="Add Subscription/Redemption", icon=WBIcon.ADD.icon, new_mode=True
            ),
        }



from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers

from wbportfolio.models import Custodian


class CustodianRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:custodian-detail")

    class Meta:
        model = Custodian
        fields = ("id", "name", "_detail")


class CustodianModelSerializer(wb_serializers.ModelSerializer):
    @wb_serializers.register_resource()
    def _addition_resources(self, instance, request, user):
        res = {
            "merge": reverse("wbportfolio:custodian-merge", args=[instance.id], request=request),
        }
        if len(instance.mapping) > 1:
            res["split"] = reverse("wbportfolio:custodian-split", args=[instance.id], request=request)
        return res

    class Meta:
        model = Custodian
        fields = ("id", "name", "mapping", "_additional_resources")



from pytest_factoryboy import register
from wbcompliance.factories.risk_management import (
    RiskIncidentTypeFactory,
    RiskRuleFactory,
    RuleBackendFactory,
    RuleThresholdFactory,
)
from wbfdm.factories import ControversyFactory
from wbportfolio.tests.conftest import *  # noqa

register(RiskIncidentTypeFactory)
register(RiskRuleFactory)
register(RuleBackendFactory)
register(RuleThresholdFactory)
register(ControversyFactory)



from .adjustments import AdjustmentTitleConfig
from .assets import (
    AssetPositionInstrumentTitleConfig,
    AssetPositionPortfolioTitleConfig,
    AssetPositionTitleConfig,
    AssetPositionUnderlyingInstrumentChartTitleConfig,
    CashPositionPortfolioTitleConfig,
    CompositionModelPortfolioPandasTitleConfig,
    ContributorPortfolioChartTitleConfig,
    DistributionChartTitleConfig,
    DistributionTableTitleConfig,
)

from .claim import (
    ConsolidatedTradeSummaryTitleConfig,
    ConsolidatedTradeSummaryDistributionChartTitleConfig,
    CumulativeNNMChartTitleConfig,
    ClaimAccountTitleConfig,
    ClaimEntryTitleConfig,
    ClaimProductTitleConfig,
    ClaimTitleConfig,
    ClaimTradeTitleConfig,
    NegativeTermimalAccountPerProductTitleConfig,
    ProfitAndLossPandasTitleConfig,
)

from .custodians import CustodianTitleConfig

from .fees import (
    FeesAggregatedProductTitleConfig,
    FeesProductTitleConfig,
    FeesTitleConfig,
)

from .instrument_prices import (
    AUMProductTitleConfig,
    InstrumentPriceAUMDataTitleConfig,
    NominalProductTitleConfig,
)

from .portfolios import PortfolioTitleConfig, DailyPortfolioCashFlowTitleConfig, PortfolioTreeGraphChartTitleConfig, TopDownPortfolioCompositionPandasTitleConfig
from .positions import (
    AggregatedAssetPositionLiquidityTitleConfig,
    AssetPositionPandasTitleConfig,
)
from .product_groups import ProductGroupTitleConfig
from .product_performance import (
    PerformanceComparisonTitleConfig,
    PerformancePandasTitleConfig,
    ProductPerformanceNetNewMoneyTitleConfig,
)
from .products import ProductPerformanceFeesTitleConfig
from .registers import RegisterTitleConfig
from .roles import PortfolioRoleInstrumentTitleConfig, PortfolioRoleTitleConfig
from .trades import (
    CustodianDistributionInstrumentTitleConfig,
    CustomerDistributionInstrumentTitleConfig,
    SubscriptionRedemptionTitleConfig,
    TradeInstrumentTitleConfig,
    TradePortfolioTitleConfig,
    TradeTitleConfig,
)
from .esg import ESGMetricAggregationPortfolioPandasTitleConfig
from .assets_and_net_new_money_progression import AssetAndNetNewMoneyProgressionChartTitleConfig



import logging
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from math import isclose
from typing import TYPE_CHECKING, Any, Iterable

import numpy as np
import pandas as pd
from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import DateRangeField
from django.db import models
from django.db.models import (
    Exists,
    F,
    OuterRef,
    Q,
    QuerySet,
    Sum,
)
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.functional import cached_property
from pandas._libs.tslibs.offsets import BDay
from skfolio.preprocessing import prices_to_returns
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.models import WBModel
from wbcore.utils.importlib import import_from_dotted_path
from wbcore.utils.models import ActiveObjectManager, DeleteToDisableMixin
from wbfdm.contrib.metric.tasks import compute_metrics_as_task
from wbfdm.enums import MarketData
from wbfdm.models import Instrument, InstrumentType
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbfdm.signals import investable_universe_updated

from wbportfolio.models.asset import AssetPosition, AssetPositionIterator
from wbportfolio.models.indexes import Index
from wbportfolio.models.portfolio_relationship import (
    InstrumentPortfolioThroughModel,
    PortfolioInstrumentPreferredClassificationThroughModel,
)
from wbportfolio.models.products import Product
from wbportfolio.pms.analytics.portfolio import Portfolio as AnalyticPortfolio
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position as PositionDTO

from . import ProductGroup
from .exceptions import InvalidAnalyticPortfolio

logger = logging.getLogger("pms")
if TYPE_CHECKING:
    from wbportfolio.models.transactions.trade_proposals import TradeProposal


def get_returns(
    instrument_ids: list[int],
    from_date: date,
    to_date: date,
    to_currency: Currency | None = None,
    ffill_returns: bool = True,
) -> tuple[dict[date, dict[int, float]], pd.DataFrame]:
    """
    Utility methods to get instrument returns for a given date range

    Args:
        from_date: date range lower bound
        to_date: date range upper bound

    Returns:
        Return a tuple of the returns and the last prices series for conveniance
    """
    # if to_currency:
    #     fx_rate = CurrencyFXRates.get_fx_rates_subquery_for_two_currencies("date", "instrument__currency", to_currency)
    # else:
    #     fx_rate = Value(Decimal(1.0))
    # prices = InstrumentPrice.objects.filter(
    #     instrument__in=instrument_ids, date__gte=from_date, date__lte=to_date
    # ).annotate(fx_rate=fx_rate, price_fx_portfolio=F("net_value") * F("fx_rate"))
    # prices_df = (
    #     pd.DataFrame(
    #         prices.filter_only_valid_prices().values_list("instrument", "price_fx_portfolio", "date"),
    #         columns=["instrument", "price_fx_portfolio", "date"],
    #     )
    #     .pivot_table(index="date", values="price_fx_portfolio", columns="instrument")
    #     .astype(float)
    #     .sort_index()
    # )
    kwargs = dict(from_date=from_date, to_date=to_date, values=[MarketData.CLOSE])
    if to_currency:
        kwargs["target_currency"] = to_currency.key
    prices_df = pd.DataFrame(Instrument.objects.filter(id__in=instrument_ids).dl.market_data(**kwargs))
    if prices_df.empty:
        raise InvalidAnalyticPortfolio()
    prices_df = prices_df[["instrument_id", "close", "valuation_date"]].pivot(
        index="valuation_date", columns="instrument_id", values="close"
    )
    ts = pd.bdate_range(prices_df.index.min(), prices_df.index.max(), freq="B")
    prices_df = prices_df.reindex(ts)
    if ffill_returns:
        prices_df = prices_df.ffill()
    prices_df.index = pd.to_datetime(prices_df.index)
    returns = prices_to_returns(prices_df, drop_inceptions_nan=False, fill_nan=ffill_returns)
    return {ts.date(): row for ts, row in prices_df.replace(np.nan, None).to_dict("index").items()}, returns.replace(
        [np.inf, -np.inf, np.nan], 0
    )


class DefaultPortfolioQueryset(QuerySet):
    def filter_invested_at_date(self, val_date: date) -> QuerySet:
        """
        Filter the queryset to get only portfolio invested at the given date
        """
        return self.filter(
            (Q(invested_timespan__startswith__lte=val_date) | Q(invested_timespan__startswith__isnull=True))
            & (Q(invested_timespan__endswith__gt=val_date) | Q(invested_timespan__endswith__isnull=True))
        )

    def filter_active_and_tracked(self):
        return self.annotate(
            has_product=Exists(
                InstrumentPortfolioThroughModel.objects.filter(
                    instrument__instrument_type=InstrumentType.PRODUCT, portfolio=OuterRef("pk")
                )
            )
        ).filter((Q(has_product=True) | Q(is_manageable=True)) & Q(is_active=True) & Q(is_tracked=True))

    def to_dependency_iterator(self, val_date: date) -> Iterable["Portfolio"]:
        """
        A method to sort the given queryset to return undependable portfolio first. This is very useful if a routine needs to be applied sequentially on portfolios by order of dependence.
        """
        MAX_ITERATIONS: int = (
            5  # in order to avoid circular dependency and infinite loop, we need to stop recursion at a max depth
        )
        remaining_portfolios = set(self)

        def _iterator(p, iterator_counter=0):
            iterator_counter += 1
            parent_portfolios = remaining_portfolios & set(
                map(lambda o: o[0], p.get_parent_portfolios(val_date))
            )  # get composition parent portfolios
            dependency_relationships = PortfolioPortfolioThroughModel.objects.filter(
                portfolio=p, dependency_portfolio__in=remaining_portfolios
            )  # get dependency portfolios
            if iterator_counter >= MAX_ITERATIONS or (
                not dependency_relationships.exists() and not bool(parent_portfolios)
            ):  # if not dependency portfolio or parent portfolio that remained, then we yield
                remaining_portfolios.remove(p)
                yield p
            else:
                # otherwise, we iterate of the dependency portfolio first
                deps_portfolios = parent_portfolios.union(
                    set([r.dependency_portfolio for r in dependency_relationships])
                )
                for deps_p in deps_portfolios:
                    yield from _iterator(deps_p, iterator_counter=iterator_counter)

        while len(remaining_portfolios) > 0:
            portfolio = next(iter(remaining_portfolios))
            yield from _iterator(portfolio)


class DefaultPortfolioManager(ActiveObjectManager):
    def get_queryset(self):
        return DefaultPortfolioQueryset(self.model).filter(is_active=True)

    def filter_invested_at_date(self, val_date: date):
        return self.get_queryset().filter_invested_at_date(val_date)

    def filter_active_and_tracked(self):
        return self.get_queryset().filter_active_and_tracked()


class ActiveTrackedPortfolioManager(DefaultPortfolioManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(asset_exists=Exists(AssetPosition.unannotated_objects.filter(portfolio=OuterRef("pk"))))
            .filter(Q(asset_exists=True) & (Q(is_tracked=True) | Q(is_manageable=True)))
        )


class PortfolioPortfolioThroughModel(models.Model):
    class Type(models.TextChoices):
        PRIMARY = "PRIMARY", "Primary"
        MODEL = "MODEL", "Model"
        CUSTODIAN = "CUSTODIAN", "Custodian"

    portfolio = models.ForeignKey("wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependency_through")
    dependency_portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependent_through"
    )
    type = models.CharField(choices=Type.choices, default=Type.PRIMARY, verbose_name="Type")

    def __str__(self):
        return f"{self.portfolio} dependant on {self.dependency_portfolio} ({self.Type[self.type].label})"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["portfolio", "type"], name="unique_primary", condition=Q(type="PRIMARY")),
            models.UniqueConstraint(fields=["portfolio", "type"], name="unique_model", condition=Q(type="MODEL")),
        ]


class Portfolio(DeleteToDisableMixin, WBModel):
    assets: models.QuerySet[AssetPosition]

    name = models.CharField(
        max_length=255,
        verbose_name="Name",
        default="",
        help_text="The Name of the Portfolio",
    )

    currency = models.ForeignKey(
        to="currency.Currency",
        related_name="portfolios",
        on_delete=models.PROTECT,
        verbose_name="Currency",
        help_text="The currency of the portfolio.",
    )
    hedged_currency = models.ForeignKey(
        to="currency.Currency",
        related_name="hedged_portfolios",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name="Hedged Currency",
        help_text="The hedged currency of the portfolio.",
    )
    depends_on = models.ManyToManyField(
        "wbportfolio.Portfolio",
        symmetrical=False,
        related_name="dependent_portfolios",
        through="wbportfolio.PortfolioPortfolioThroughModel",
        through_fields=("portfolio", "dependency_portfolio"),
        blank=True,
        verbose_name="The portfolios this portfolio depends on",
    )

    preferred_instrument_classifications = models.ManyToManyField(
        "wbfdm.Instrument",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
        related_name="preferred_portfolio_classifications",
        through="wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel",
        through_fields=("portfolio", "instrument"),
        blank=True,
        verbose_name="The Preferred classification per instrument",
    )
    instruments = models.ManyToManyField(
        "wbfdm.Instrument",
        through=InstrumentPortfolioThroughModel,
        related_name="portfolios",
        blank=True,
        verbose_name="Instruments",
        help_text="Instruments linked to this instrument",
    )
    invested_timespan = DateRangeField(
        null=True, blank=True, help_text="Define when this portfolio is considered invested"
    )

    is_manageable = models.BooleanField(
        default=False,
        help_text="True if the portfolio can be manually modified (e.g. Trade proposal be submitted or total weight recomputed)",
    )
    is_tracked = models.BooleanField(
        default=True,
        help_text="True if the internal updating mechanism (e.g., Next weights or Look-Through computation, rebalancing etc...) needs to apply to this portfolio",
    )
    only_weighting = models.BooleanField(
        default=False,
        help_text="Indicates that this portfolio is only utilizing weights and disregards shares, e.g. a model portfolio",
    )
    is_lookthrough = models.BooleanField(
        default=False,
        help_text="Indicates that this portfolio is a look-through portfolio",
    )
    is_composition = models.BooleanField(
        default=False, help_text="If true, this portfolio is a composition of other portfolio"
    )
    updated_at = models.DateTimeField(blank=True, null=True, verbose_name="Updated At")
    last_position_date = models.DateField(blank=True, null=True, verbose_name="Last Position Date")
    initial_position_date = models.DateField(blank=True, null=True, verbose_name="Last Position Date")

    bank_accounts = models.ManyToManyField(
        to="directory.BankingContact",
        related_name="wbportfolio_portfolios",
        through="wbportfolio.PortfolioBankAccountThroughModel",
        blank=True,
    )

    objects = DefaultPortfolioManager()
    tracked_objects = ActiveTrackedPortfolioManager()

    @property
    def primary_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.PRIMARY
            ).dependency_portfolio

    @property
    def model_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
            ).dependency_portfolio

    @property
    def composition_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self,
                type=PortfolioPortfolioThroughModel.Type.MODEL,
                dependency_portfolio__is_composition=True,
            ).dependency_portfolio

    @property
    def imported_assets(self):
        return self.assets.filter(is_estimated=False)

    @cached_property
    def pms_instruments(self):
        instruments = [i for i in Product.objects.filter(portfolios=self)]
        instruments.extend([i for i in ProductGroup.objects.filter(portfolios=self)])
        instruments.extend([i for i in Index.objects.filter(portfolios=self)])
        return instruments

    @property
    def can_be_rebalanced(self):
        return self.is_manageable and not self.is_lookthrough

    def delete(self, **kwargs):
        super().delete(**kwargs)
        # We check if for all linked instruments, this portfolio was the last active one (if yes, we disable the instrument)
        if self.id:
            for instrument in self.instruments.iterator():
                if not instrument.portfolios.filter(is_active=True).exists():
                    instrument.delisted_date = date.today() - timedelta(days=1)
                    instrument.save()

    def _build_dto(self, val_date: date, include_drift_factor: bool = False, **extra_kwargs) -> PortfolioDTO:
        "returns the dto representation of this portfolio at the specified date"
        assets = self.assets.filter(date=val_date, **extra_kwargs)
        try:
            drifted_weights = self.get_analytic_portfolio(val_date).get_next_weights()
        except InvalidAnalyticPortfolio:
            drifted_weights = {}
        positions = []
        for asset in assets:
            drift_factor = (
                Decimal(drifted_weights.get(asset.underlying_quote.id, asset.weighting)) / asset.weighting
                if asset.weighting
                else Decimal(1.0)
            )
            positions.append(asset._build_dto(drift_factor=drift_factor))

        return PortfolioDTO(positions)

    def get_weights(self, val_date: date) -> dict[int, float]:
        """
        A convenience utility method to returns the portfolio weights for this portfolio as a dictionary (instrument id as key and weights as value)

        Args:
            val_date: The date at which to return the weights for this portfolio

        Returns:
            A dictionary containing the weights for this portfolio
        """
        return dict(
            map(
                lambda r: (r[0], float(r[1])),
                self.assets.filter(date=val_date)
                .values("underlying_quote")
                .annotate(sum_weight=Sum("weighting"))
                .values_list("underlying_quote", "sum_weight"),
            )
        )

    def get_analytic_portfolio(
        self, val_date: date, weights: dict[int, float] | None = None, **kwargs
    ) -> AnalyticPortfolio:
        """
        Return the analytic portfolio associated with this portfolio at the given date

        the analytic portfolio inherit from SKFolio Portfolio and can be used to access all this library methods
        Args:
            val_date: the date to calculate the portfolio for

        Returns:
            The instantiated analytic portfolio
        """
        if not weights:
            weights = self.get_weights(val_date)
        return_date = (val_date + BDay(1)).date()
        _, returns = get_returns(
            list(weights.keys()), (val_date - BDay(2)).date(), return_date, to_currency=self.currency, **kwargs
        )
        if pd.Timestamp(return_date) not in returns.index:
            raise InvalidAnalyticPortfolio()
        returns = returns.fillna(0)  # not sure this is what we want
        return AnalyticPortfolio(
            X=returns,
            weights=weights,
        )

    def is_invested_at_date(self, val_date: date) -> bool:
        return (
            self.invested_timespan
            and self.invested_timespan.upper > val_date
            and self.invested_timespan.lower <= val_date
        )

    def __str__(self):
        return f"{self.id:06}: {self.name}"

    class Meta:
        verbose_name = "Portfolio"
        verbose_name_plural = "Portfolios"

        notification_types = [
            create_notification_type(
                "wbportfolio.portfolio.check_custodian_portfolio",
                "Check Custodian Portfolio",
                "Sends a notification when a portfolio does not match with its custodian portfolio",
                True,
                True,
                True,
            ),
            create_notification_type(
                "wbportfolio.portfolio.replay_done",
                "Portfolio Replay finished",
                "Sends a notification when a the requested trade proposal replay is done",
                True,
                True,
                True,
            ),
        ]

    def is_active_at_date(self, val_date: date) -> bool:
        """
        Return if the base instrument has a total aum greater than 0
        :val_date: the date at which we need to evaluate if the portfolio is considered active
        """
        active_portfolio = self.is_active or self.deletion_datetime.date() > val_date
        if self.instruments.exists():
            return active_portfolio and any(
                [instrument.is_active_at_date(val_date) for instrument in self.instruments.all()]
            )
        return active_portfolio

    def get_total_asset_value(self, val_date: date) -> Decimal:
        """
        Return the total asset under management of the portfolio at the specified valuation date
        Args:
            val_date: The date at which aum needs to be computed
        Returns:
            The total AUM (0 if there is no position)
        """
        return self.assets.filter(date=val_date).aggregate(s=Sum("total_value_fx_portfolio"))["s"] or Decimal(0.0)

    def get_total_asset_under_management(self, val_date):
        from wbportfolio.models.transactions.trades import Trade

        trades = Trade.valid_customer_trade_objects.filter(portfolio=self, transaction_date__lte=val_date)

        total_aum = Decimal(0)
        for underlying_instrument_id, sum_shares in (
            trades.values("underlying_instrument")
            .annotate(
                sum_shares=Sum("shares"),
            )
            .values_list("underlying_instrument", "sum_shares")
        ):
            with suppress(Instrument.DoesNotExist, InstrumentPrice.DoesNotExist):
                instrument = Instrument.objects.get(id=underlying_instrument_id)
                last_price = instrument.valuations.filter(date__lte=val_date).latest("date").net_value
                fx_rate = instrument.currency.convert(val_date, self.currency)
                total_aum += last_price * sum_shares * fx_rate
        return total_aum

    def _get_assets(self, with_estimated=True, with_cash=True):
        qs = self.assets
        if not with_estimated:
            qs = qs.filter(is_estimated=False)
        if not with_cash:
            qs = qs.exclude(underlying_instrument__is_cash=True)
        return qs

    def get_earliest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__gte=val_date)
        if qs.exists():
            return qs.earliest("date").date
        return None

    def get_latest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__lte=val_date)

        if qs.exists():
            return qs.latest("date").date
        return None

    # Asset Position Utility Functions
    def get_holding(self, val_date, exclude_cash=True, exclude_index=True):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date, weighting__gt=0)
        if exclude_index:
            qs = qs.exclude(underlying_instrument__instrument_type=InstrumentType.INDEX)
        return (
            qs.values("underlying_instrument__name")
            .annotate(total_value_fx_portfolio=Sum("total_value_fx_portfolio"), weighting=Sum("weighting"))
            .order_by("-total_value_fx_portfolio")
        )

    def _get_groupedby_df(
        self,
        group_by,
        val_date: date,
        exclude_cash: bool | None = False,
        exclude_index: bool | None = False,
        extra_filter_parameters: dict[str, Any] = None,
        **groupby_kwargs,
    ):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date)
        if exclude_index:
            # We exclude only index that are not considered as cash. Setting exclude_cash to true convers this case.
            qs = qs.exclude(
                Q(underlying_instrument__instrument_type=InstrumentType.INDEX)
                & Q(underlying_instrument__is_cash=False)
            )
        if extra_filter_parameters:
            qs = qs.filter(**extra_filter_parameters)
        qs = group_by(qs, **groupby_kwargs).annotate(sum_weighting=Sum(F("weighting"))).order_by("-sum_weighting")
        df = pd.DataFrame(
            qs.values_list("aggregated_title", "sum_weighting"), columns=["aggregated_title", "weighting"]
        )
        if not df.empty:
            df.weighting = df.weighting.astype("float")
            df.weighting = df.weighting / df.weighting.sum()
            df = df.sort_values(by=["weighting"])
        return df.where(pd.notnull(df), None)

    def get_geographical_breakdown(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.country_group_by, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_currency_exposure(self, val_date, **kwargs):
        df = self._get_groupedby_df(AssetPosition.currency_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_market_cap_distribution(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.marketcap_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_instrument__instrument_type": InstrumentType.EQUITY},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_liquidity(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.liquidity_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_instrument__instrument_type": InstrumentType.EQUITY},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_industry_exposure(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.group_by_primary, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_asset_allocation(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(AssetPosition.cash_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_adjusted_child_positions(self, val_date):
        if (
            child_positions := self.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        ).count() == 1:
            if portfolio := child_positions.first().underlying_instrument.primary_portfolio:
                child_positions = portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        for position in child_positions:
            if child_portfolio := position.underlying_instrument.primary_portfolio:
                index_positions = child_portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(
                    date=val_date
                )

                for index_position in index_positions.all():
                    weighting = index_position.weighting * position.weighting
                    if weighting != 0:
                        yield {
                            "underlying_instrument_id": index_position.underlying_instrument.id,
                            "weighting": weighting,
                        }

    def get_longshort_distribution(self, val_date):
        df = pd.DataFrame(self.get_adjusted_child_positions(val_date))

        if not df.empty:
            df["is_cash"] = df.underlying_instrument_id.apply(lambda x: Instrument.objects.get(id=x).is_cash)
            df = df[~df["is_cash"]]
            df = (
                df[["underlying_instrument_id", "weighting"]].groupby("underlying_instrument_id").sum().astype("float")
            )
            df.weighting = df.weighting / df.weighting.sum()
            short_weight = df[df.weighting < 0].weighting.abs().sum()
            long_weight = df[df.weighting > 0].weighting.sum()
            total_weight = long_weight + short_weight
            return pd.DataFrame(
                [
                    {"title": "Long", "weighting": long_weight / total_weight},
                    {"title": "Short", "weighting": short_weight / total_weight},
                ]
            )
        return df

    def get_portfolio_contribution_df(
        self,
        start: date,
        end: date,
        with_cash: bool = True,
        hedged_currency: Currency | None = None,
        only_equity: bool = False,
    ) -> pd.DataFrame:
        qs = self._get_assets(with_cash=with_cash).filter(date__gte=start, date__lte=end)
        if only_equity:
            qs = qs.filter(underlying_instrument__instrument_type=InstrumentType.EQUITY)
        qs = qs.annotate_hedged_currency_fx_rate(hedged_currency)
        df = Portfolio.get_contribution_df(
            qs.select_related("underlying_instrument").values_list(
                "date", "price", "hedged_currency_fx_rate", "underlying_instrument", "weighting"
            )
        )
        df = df.rename(columns={"group_key": "underlying_instrument"})
        df["underlying_instrument__name_repr"] = df["underlying_instrument"].map(
            dict(Instrument.objects.filter(id__in=df["underlying_instrument"]).values_list("id", "name_repr"))
        )
        return df

    def check_related_portfolio_at_date(self, val_date: date, related_portfolio: "Portfolio"):
        assets = AssetPosition.objects.filter(
            date=val_date, underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ).values("underlying_instrument", "shares")
        assets1 = assets.filter(portfolio=self)
        assets2 = assets.filter(portfolio=related_portfolio)
        return assets1.difference(assets2)

    def get_child_portfolios(self, val_date: date) -> set["Portfolio"]:
        child_portfolios = set()
        instrument_rel = InstrumentPortfolioThroughModel.objects.filter(portfolio=self)
        if instrument_rel.exists():
            for parent_portfolio in Portfolio.objects.filter(
                id__in=AssetPosition.unannotated_objects.filter(
                    date=val_date, underlying_quote__in=instrument_rel.values("instrument")
                ).values("portfolio")
            ):
                child_portfolios.add(parent_portfolio)
        return child_portfolios

    def get_parent_portfolios(self, val_date: date) -> set["Portfolio"]:
        for asset in self.assets.filter(date=val_date, underlying_instrument__portfolios__isnull=False).distinct(
            "underlying_instrument"
        ):
            if portfolio := asset.underlying_instrument.portfolio:
                yield portfolio, asset.weighting

    def get_next_rebalancing_date(self, start_date: date) -> date | None:
        if automatic_rebalancer := getattr(self, "automatic_rebalancer", None):
            return automatic_rebalancer.get_next_rebalancing_date(start_date)

    def change_at_date(
        self,
        val_date: date,
        recompute_weighting: bool = False,
        force_recompute_weighting: bool = False,
        evaluate_rebalancer: bool = True,
        changed_weights: dict[int, float] | None = None,
    ):
        logger.info(f"change at date for {self} at {val_date}")

        if recompute_weighting:
            # We normalize weight across the portfolio for a given date
            qs = self.assets.filter(date=val_date).filter(
                Q(total_value_fx_portfolio__isnull=False) | Q(weighting__isnull=False)
            )
            if (self.is_lookthrough or self.is_manageable or force_recompute_weighting) and qs.exists():
                total_weighting = qs.aggregate(s=Sum("weighting"))["s"]
                # We check if this actually necessary
                # (i.e. if the weight is already summed to 100%, it is already normalized)
                if (
                    not total_weighting
                    or not isclose(total_weighting, Decimal(1.0), abs_tol=0.001)
                    or force_recompute_weighting
                ):
                    total_value = qs.aggregate(s=Sum("total_value_fx_portfolio"))["s"]
                    # TODO we change this because postgres doesn't support join statement in update (and total_value_fx_portfolio is a joined annoted field)
                    for asset in qs:
                        if total_value:
                            asset.weighting = asset._total_value_fx_portfolio / total_value
                        elif total_weighting:
                            asset.weighting = asset.weighting / total_weighting
                        asset.save()

        # We check if there is an instrument attached to the portfolio with calculated NAV and price computation method
        self.estimate_net_asset_values(
            (val_date + BDay(1)).date(), weights=changed_weights
        )  # updating weighting in t0 influence nav in t+1
        if evaluate_rebalancer:
            self.evaluate_rebalancing(val_date)

        self.updated_at = timezone.now()
        if self.assets.filter(date=val_date).exists():
            if not self.last_position_date or self.last_position_date < val_date:
                self.last_position_date = val_date
            if not self.initial_position_date or self.initial_position_date > val_date:
                self.initial_position_date = val_date
        self.save()

        self.handle_controlling_portfolio_change_at_date(val_date)

    def handle_controlling_portfolio_change_at_date(self, val_date: date):
        if self.is_tracked:
            for rel in PortfolioPortfolioThroughModel.objects.filter(
                dependency_portfolio=self,
                type=PortfolioPortfolioThroughModel.Type.PRIMARY,
                portfolio__is_lookthrough=True,
            ):
                rel.portfolio.compute_lookthrough(val_date)
            for rel in PortfolioPortfolioThroughModel.objects.filter(
                dependency_portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
            ):
                rel.portfolio.evaluate_rebalancing(val_date)
            for dependent_portfolio in self.get_child_portfolios(val_date):
                dependent_portfolio.change_at_date(val_date)
                dependent_portfolio.handle_controlling_portfolio_change_at_date(val_date)

    def evaluate_rebalancing(self, val_date: date):
        if hasattr(self, "automatic_rebalancer"):
            # if the portfolio has an automatic rebalancer and the next business day is suitable with the rebalancer, we create a trade proposal automatically
            next_business_date = (val_date + BDay(1)).date()
            if self.automatic_rebalancer.is_valid(val_date):  # we evaluate the rebalancer in t0 and t+1
                logger.info(f"Evaluate Rebalancing for {self} at {val_date}")
                self.automatic_rebalancer.evaluate_rebalancing(val_date)
            if self.automatic_rebalancer.is_valid(next_business_date):
                logger.info(f"Evaluate Rebalancing for {self} at {next_business_date}")
                self.automatic_rebalancer.evaluate_rebalancing(next_business_date)

    def estimate_net_asset_values(self, val_date: date, weights: dict[int | float] | None = None):
        for instrument in self.pms_instruments:
            if instrument.is_active_at_date(val_date) and (
                net_asset_value_computation_method_path := instrument.net_asset_value_computation_method_path
            ):
                logger.info(f"Estimate NAV of {val_date:%Y-%m-%d} for instrument {instrument}")
                net_asset_value_computation_method = import_from_dotted_path(net_asset_value_computation_method_path)
                estimated_net_asset_value = net_asset_value_computation_method(val_date, instrument, weights=weights)
                if estimated_net_asset_value is not None:
                    InstrumentPrice.objects.update_or_create(
                        instrument=instrument,
                        date=val_date,
                        calculated=True,
                        defaults={
                            "gross_value": estimated_net_asset_value,
                            "net_value": estimated_net_asset_value,
                        },
                    )
                    if (
                        val_date == instrument.last_price_date
                    ):  # if price date is the latest instrument price date, we recompute the last valuation data
                        instrument.update_last_valuation_date()

    def drift_weights(
        self, start_date: date, end_date: date, stop_at_rebalancing: bool = False
    ) -> tuple[AssetPositionIterator, "TradeProposal"]:
        logger.info(f"drift weights for {self} from {start_date:%Y-%m-%d} to {end_date:%Y-%m-%d}")
        rebalancer = getattr(self, "automatic_rebalancer", None)
        # Get initial weights
        weights = self.get_weights(start_date)  # initial weights
        if not weights:
            previous_date = self.assets.filter(date__lte=start_date).latest("date").date
            drifted_positions, _ = self.drift_weights(previous_date, start_date)
            weights = drifted_positions.get_weights()[start_date]

        # Get returns and prices data for the whole date range
        instrument_ids = list(weights.keys())

        prices, returns = get_returns(
            instrument_ids,
            (start_date - BDay(3)).date(),
            end_date,
            to_currency=self.currency,
            ffill_returns=True,
        )
        # Get raw prices to speed up asset position creation
        # Instantiate the position iterator with the initial weights
        positions = AssetPositionIterator(self, prices=prices)
        last_trade_proposal = None
        for to_date_ts in pd.date_range(start_date + timedelta(days=1), end_date, freq="B"):
            to_date = to_date_ts.date()
            to_is_active = self.is_active_at_date(to_date)
            logger.info(f"Processing {to_date:%Y-%m-%d}")

            try:
                last_returns = returns.loc[[to_date_ts], :]
                analytic_portfolio = AnalyticPortfolio(weights=weights, X=last_returns)
                drifted_weights = analytic_portfolio.get_next_weights()
            except KeyError:  # if no return for that date, we break and continue
                drifted_weights = weights

            if rebalancer and rebalancer.is_valid(to_date):
                effective_portfolio = PortfolioDTO(
                    positions=[
                        PositionDTO(
                            date=to_date,
                            underlying_instrument=i,
                            weighting=Decimal(w),
                            drift_factor=Decimal(drifted_weights.get(i, w) / float(w)) if w else Decimal(1.0),
                        )
                        for i, w in weights.items()
                    ]
                )
                last_trade_proposal = rebalancer.evaluate_rebalancing(to_date, effective_portfolio=effective_portfolio)
                if stop_at_rebalancing:
                    break
                next_weights = {
                    trade.underlying_instrument: float(trade._target_weight)
                    for trade in last_trade_proposal.trades.all().annotate_base_info()
                }
                positions.add((to_date, next_weights), is_estimated=False)
            else:
                next_weights = drifted_weights
                if to_is_active:
                    positions.add((to_date, next_weights))
                else:
                    positions.add(
                        (to_date, {underlying_quote_id: 0.0 for underlying_quote_id in weights.keys()})
                    )  # if we have no return or portfolio is not active anymore, we return an emptied portfolio
                    break
            weights = next_weights
        return positions, last_trade_proposal

    def propagate_or_update_assets(self, from_date: date, to_date: date):
        """
        Create a new portfolio at `to_date` based on the portfolio in `from_date`.

        Args:
            from_date: The date to propagate the portfolio from
            to_date:  The date to create the new portfolio at

        """
        # we don't propagate on already imported portfolio by default
        is_target_portfolio_imported = self.assets.filter(date=to_date, is_estimated=False).exists()
        if (
            not self.is_lookthrough and not is_target_portfolio_imported and self.is_active_at_date(from_date)
        ):  # we cannot propagate a new portfolio for untracked, or look-through or already imported or inactive portfolios
            positions, _ = self.drift_weights(from_date, to_date)
            self.bulk_create_positions(
                positions, delete_leftovers=True, compute_metrics=True, evaluate_rebalancer=False
            )

    def get_lookthrough_positions(
        self,
        sync_date: date,
        portfolio_total_asset_value: Decimal | None = None,
        with_intermediary_position: bool = False,
    ):
        """Recursively calculates the look-through position for a portfolio

        Arguments:
            sync_date {datetime.date} -- The date on which the assets will be computed
            portfolio_total_value: {Decimal} -- The total value of the portfolio (needed to compute initial shares)
        """

        def _crawl_portfolio(
            parent_portfolio,
            adjusted_weighting,
            adjusted_currency_fx_rate,
            adjusted_is_estimated,
            path=None,
        ):
            if not path:
                path = []
            path.append(parent_portfolio)
            for position in parent_portfolio.assets.filter(date=sync_date):
                position.id = None
                position.weighting = adjusted_weighting * position.weighting
                position.initial_currency_fx_rate = adjusted_currency_fx_rate * position.currency_fx_rate
                position.is_estimated = (adjusted_is_estimated or position.is_estimated) and not (
                    position.weighting == 1.0
                )
                # to get from which portfolio this position is created, we need to differantiate between:
                # * Composition portfolio: where the portfolio created is the second encountered portfolio
                # * Other: portfolio created is the last encountered portfolio
                # If `path` is empty, we use None as portfolio_created
                try:
                    if self.is_composition:
                        position.portfolio_created = path[1]
                    else:
                        position.portfolio_created = path[-1]
                except IndexError:
                    position.portfolio_created = None

                setattr(position, "path", path)
                position.initial_shares = None
                if portfolio_total_asset_value and (price_fx_portfolio := position.price * position.currency_fx_rate):
                    position.initial_shares = (position.weighting * portfolio_total_asset_value) / price_fx_portfolio
                if child_portfolio := position.underlying_quote.primary_portfolio:
                    if with_intermediary_position:
                        yield position
                    yield from _crawl_portfolio(
                        child_portfolio,
                        position.weighting,
                        position.currency_fx_rate,
                        position.is_estimated,
                        path=path.copy(),
                    )
                elif position.weighting:  # we do not yield position with weight 0 because of issue with certain multi-thematic portfolios which contain duplicates
                    yield position

        yield from _crawl_portfolio(self, Decimal(1.0), Decimal(1.0), False)

    def get_positions(self, val_date: date, **kwargs) -> Iterable[AssetPosition]:
        if self.is_composition:
            assets = list(self.get_lookthrough_positions(val_date, **kwargs))
        else:
            assets = list(self.assets.filter(date=val_date))
        return assets

    def compute_lookthrough(self, from_date: date, to_date: date | None = None):
        if not self.primary_portfolio or not self.is_lookthrough:
            raise ValueError(
                "Lookthrough position can only be computed on lookthrough portfolio with a primary portfolio"
            )
        positions = AssetPositionIterator(self)
        if not to_date:
            to_date = from_date
        for from_date in pd.date_range(from_date, to_date, freq="B").date:
            logger.info(f"Compute Look-Through for {self} at {from_date}")
            portfolio_total_asset_value = (
                self.primary_portfolio.get_total_asset_under_management(from_date) if not self.only_weighting else None
            )
            positions.add(
                list(self.primary_portfolio.get_lookthrough_positions(from_date, portfolio_total_asset_value)),
            )
        self.bulk_create_positions(positions, delete_leftovers=True, compute_metrics=True)

    def update_preferred_classification_per_instrument(self):
        # Function to automatically assign Preferred instrument based on the assets' underlying instruments of the
        # attached wbportfolio
        instruments = filter(
            None,
            map(
                lambda x: Instrument.objects.get(id=x["underlying_instrument"]).get_classifable_ancestor(
                    include_self=True
                ),
                self.assets.values("underlying_instrument").distinct("underlying_instrument"),
            ),
        )
        leftovers_instruments = list(
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(portfolio=self).values_list(
                "instrument", flat=True
            )
        )
        for instrument in instruments:
            other_classifications = instrument.classifications.filter(group__is_primary=False)
            default_classification = None
            if other_classifications.count() == 1:
                default_classification = other_classifications.first()
            if not PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument
            ).exists():
                PortfolioInstrumentPreferredClassificationThroughModel.objects.create(
                    portfolio=self,
                    instrument=instrument,
                    classification=default_classification,
                    classification_group=default_classification.group if default_classification else None,
                )
            if instrument.id in leftovers_instruments:
                leftovers_instruments.remove(instrument.id)

        for instrument_id in leftovers_instruments:
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument_id
            ).delete()

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfolio"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:portfoliorepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    def bulk_create_positions(
        self,
        positions: AssetPositionIterator,
        delete_leftovers: bool = False,
        force_save: bool = False,
        compute_metrics: bool = True,
        broadcast_changes_at_date: bool = True,
        **kwargs,
    ):
        if positions:
            # we need to delete the existing estimated portfolio because otherwise we risk to have existing and not
            # overlapping positions remaining (as they will not be updating by the bulk create). E.g. when someone
            # change completely the trades of a portfolio model and drift it.

            dates = positions.get_dates()
            self.assets.filter(date__in=dates, is_estimated=True).delete()

            if self.is_tracked or force_save:  # if the portfolio is not "tracked", we do no drift weights
                leftover_positions_ids = list(
                    self.assets.filter(date__in=dates).values_list("id", flat=True)
                )  # we need to get the ids otherwise the queryset is reevaluated later
                positions_list = list(positions)
                logger.info(
                    f"bulk saving {len(positions_list)} positions ({len(leftover_positions_ids)} leftovers) ..."
                )
                objs = AssetPosition.unannotated_objects.bulk_create(
                    positions_list,
                    update_fields=[
                        "weighting",
                        "initial_price",
                        "initial_currency_fx_rate",
                        "initial_shares",
                        "currency_fx_rate_instrument_to_usd",
                        "currency_fx_rate_portfolio_to_usd",
                        "underlying_quote_price",
                        "portfolio",
                        "portfolio_created",
                        "underlying_instrument",
                    ],
                    unique_fields=["portfolio", "date", "underlying_quote", "portfolio_created"],
                    update_conflicts=True,
                    batch_size=10000,
                )
                if delete_leftovers:
                    objs_ids = list(map(lambda x: x.id, objs))
                    leftover_positions_ids = list(filter(lambda i: i not in objs_ids, leftover_positions_ids))
                    logger.info(f"deleting {len(leftover_positions_ids)} leftover positions..")
                    AssetPosition.objects.filter(id__in=leftover_positions_ids).delete()
            if compute_metrics and self.is_tracked:
                for val_date in dates:
                    compute_metrics_as_task.delay(
                        val_date,
                        basket_id=self.id,
                        basket_content_type_id=ContentType.objects.get_for_model(Portfolio).id,
                    )
            if broadcast_changes_at_date:
                for update_date, changed_weights in positions.get_weights().items():
                    self.change_at_date(update_date, changed_weights=changed_weights, **kwargs)

    @classmethod
    def _get_or_create_portfolio(cls, instrument_handler, portfolio_data):
        if isinstance(portfolio_data, int):
            return Portfolio.all_objects.get(id=portfolio_data)
        instrument = portfolio_data
        if isinstance(portfolio_data, dict):
            instrument = instrument_handler.process_object(instrument, only_security=False, read_only=True)[0]
        return instrument.primary_portfolio

    def check_share_diff(self, val_date: date) -> bool:
        return self.assets.filter(Q(date=val_date) & ~Q(initial_shares=F("initial_shares_at_custodian"))).exists()

    @classmethod
    def get_contribution_df(cls, data, need_normalize: bool = False):
        df = pd.DataFrame(
            data,
            columns=[
                "date",
                "price",
                "currency_fx_rate",
                "group_key",
                "value",
            ],
        )
        if not df.empty:
            df = df[df["value"] != 0]
            df.date = pd.to_datetime(df.date)
            df["price_fx_portfolio"] = df.price * df.currency_fx_rate

            df[["price", "price_fx_portfolio", "value", "currency_fx_rate"]] = df[
                ["price", "price_fx_portfolio", "value", "currency_fx_rate"]
            ].astype("float")

            df["group_key"] = df["group_key"].fillna(0)

            df = (
                df[
                    [
                        "group_key",
                        "date",
                        "price",
                        "price_fx_portfolio",
                        "value",
                        "currency_fx_rate",
                    ]
                ]
                .groupby(["date", "group_key"], dropna=False)
                .agg(
                    {
                        "price": "mean",
                        "price_fx_portfolio": "mean",
                        "value": "sum",
                        "currency_fx_rate": "mean",
                    }
                )
                .reset_index()
                .set_index("date")
                .sort_index()
            )
            df["value"] = df["value"].fillna(0)
            value = df.pivot_table(
                index="date",
                columns=["group_key"],
                values="value",
                fill_value=0,
                aggfunc="sum",
            )
            weights_ = value
            if need_normalize:
                total_value_price = df["value"].groupby("date", dropna=False).sum()
                weights_ = value.divide(total_value_price, axis=0)
            prices_usd = (
                df.pivot_table(
                    index="date",
                    columns=["group_key"],
                    values="price_fx_portfolio",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            rates_fx = (
                df.pivot_table(
                    index="date",
                    columns=["group_key"],
                    values="currency_fx_rate",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            prices_usd = prices_usd.ffill()
            performance_prices = prices_usd / prices_usd.shift(1, axis=0) - 1
            contributions_prices = performance_prices.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_prices = (1 + contributions_prices.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_prices = contributions_prices.multiply(total_contrib_prices, axis=0).sum(skipna=False)
            monthly_perf_prices = (1 + performance_prices).dropna(how="all").product(axis=0, skipna=False) - 1

            rates_fx = rates_fx.ffill()
            performance_rates_fx = rates_fx / rates_fx.shift(1, axis=0) - 1
            contributions_rates_fx = performance_rates_fx.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_rates_fx = (1 + contributions_rates_fx.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_rates_fx = contributions_rates_fx.multiply(total_contrib_rates_fx, axis=0).sum(skipna=False)
            monthly_perf_rates_fx = (1 + performance_rates_fx).dropna(how="all").product(axis=0, skipna=False) - 1

            res = pd.concat(
                [
                    monthly_perf_prices,
                    monthly_perf_rates_fx,
                    contributions_prices,
                    contributions_rates_fx,
                    weights_.iloc[0, :],
                    weights_.iloc[-1, :],
                    value.iloc[0, :],
                    value.iloc[-1, :],
                ],
                axis=1,
            ).reset_index()
            res.columns = [
                "group_key",
                "performance_total",
                "performance_forex",
                "contribution_total",
                "contribution_forex",
                "allocation_start",
                "allocation_end",
                "total_value_start",
                "total_value_end",
            ]

            return res.replace([np.inf, -np.inf, np.nan], 0)
        return pd.DataFrame()

    def get_or_create_index(self):
        index = Index.objects.create(name=self.name, currency=self.currency)
        index.portfolios.all().delete()
        InstrumentPortfolioThroughModel.objects.update_or_create(instrument=index, defaults={"portfolio": self})

    @classmethod
    def create_model_portfolio(cls, name: str, currency: Currency, with_index: bool = True):
        portfolio = cls.objects.create(
            is_manageable=True,
            name=name,
            currency=currency,
        )
        if with_index:
            portfolio.get_or_create_index()
        return portfolio


def default_estimate_net_value(
    val_date: date, instrument: Instrument, weights: dict[int, float] | None = None
) -> float | None:
    portfolio: Portfolio = instrument.portfolio
    previous_val_date = (val_date - BDay(1)).date()
    if not weights:
        weights = portfolio.get_weights(previous_val_date)
    # we assume that in t-1 we will have a portfolio (with at least estimate position). If we use the latest position date before val_date, we run into the problem of being able to compute nav at every date
    if weights and (last_price := instrument.get_latest_price(previous_val_date)):
        with suppress(
            IndexError, InvalidAnalyticPortfolio
        ):  # we silent any indexerror introduced by no returns for the past days
            analytic_portfolio = portfolio.get_analytic_portfolio(previous_val_date, weights=weights)
            return analytic_portfolio.get_estimate_net_value(float(last_price.net_value))


@receiver(post_save, sender="wbportfolio.PortfolioPortfolioThroughModel")
def post_portfolio_relationship_creation(sender, instance, created, raw, **kwargs):
    if (
        not raw
        and created
        and instance.portfolio.is_lookthrough
        and instance.type == PortfolioPortfolioThroughModel.Type.PRIMARY
    ):
        with suppress(AssetPosition.DoesNotExist):
            earliest_primary_position_date = instance.dependency_portfolio.assets.earliest("date").date
            compute_lookthrough_as_task.delay(instance.portfolio.id, earliest_primary_position_date, date.today())


@shared_task(queue="portfolio")
def trigger_portfolio_change_as_task(portfolio_id, val_date, **kwargs):
    portfolio = Portfolio.all_objects.get(id=portfolio_id)
    portfolio.change_at_date(val_date, **kwargs)


@shared_task(queue="portfolio")
def compute_lookthrough_as_task(portfolio_id: int, start: date, end: date):
    portfolio = Portfolio.objects.get(id=portfolio_id)
    portfolio.compute_lookthrough(start, to_date=end)


@receiver(investable_universe_updated, sender="wbfdm.Instrument")
def update_portfolio_after_investable_universe(*args, end_date: date | None = None, **kwargs):
    if not end_date:
        end_date = date.today()
    end_date = (end_date + timedelta(days=1) - BDay(1)).date()  # shift in case of business day
    from_date = (end_date - BDay(1)).date()
    for portfolio in Portfolio.tracked_objects.all().to_dependency_iterator(from_date):
        if not portfolio.is_lookthrough:
            try:
                portfolio.propagate_or_update_assets(from_date, end_date)
            except Exception as e:
                logger.error(f"Exception while propagating portfolio assets {portfolio}: {e}")
        portfolio.estimate_net_asset_values(end_date)



from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbfdm.serializers import InvestableUniverseRepresentationSerializer

from wbportfolio.models import DividendTransaction
from wbportfolio.serializers import PortfolioRepresentationSerializer


class DividendRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = DividendTransaction
        fields = ("id", "value_date")


class DividendModelSerializer(wb_serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    _underlying_instrument = InvestableUniverseRepresentationSerializer(source="underlying_instrument")
    _currency = CurrencyRepresentationSerializer(source="currency")

    total_value = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True, required=False)
    total_value_fx_portfolio = wb_serializers.DecimalField(
        max_digits=14, decimal_places=2, read_only=True, required=False
    )
    total_value_gross = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True, required=False)
    total_value_gross_fx_portfolio = wb_serializers.DecimalField(
        max_digits=14, decimal_places=2, read_only=True, required=False
    )

    class Meta:
        model = DividendTransaction
        fields = (
            "id",
            "value_date",
            "record_date",
            "ex_date",
            "portfolio",
            "underlying_instrument",
            "currency",
            "_portfolio",
            "_underlying_instrument",
            "_currency",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
        )



def register(model_name: str):
    """
    Decorator to include when a backend need automatic registration
    """
    from wbportfolio.models.transactions.rebalancing import RebalancingModel

    def _decorator(backend_class):
        defaults = {
            "name": model_name,
        }
        RebalancingModel.objects.update_or_create(
            class_path=backend_class.__module__ + "." + backend_class.__name__,
            defaults=defaults,
        )
        return backend_class

    return _decorator



from wbcore.metadata.configs.titles import TitleViewConfig


class ClaimTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Claims"

    def get_instance_title(self):
        return "{{_claimant.computed_str}}: {{shares}} of {{_product.name}} ({{bank}} - {{date}})"

    def get_create_title(self):
        return "New Claim"


class ClaimProductTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"Claims for Product {self.view.product.computed_str}"

    def get_instance_title(self):
        return f"Claim for Product {self.view.product.computed_str}"

    def get_create_title(self):
        return f"New Claim for Product {self.view.product.computed_str}"


class ClaimAccountTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"Claims for Account {self.view.account.computed_str}"

    def get_instance_title(self):
        return f"Claim for Account {self.view.account.computed_str}"

    def get_create_title(self):
        return f"New Claim for Account {self.view.account.computed_str}"


class ClaimEntryTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"Claims from {self.view.entry.computed_str}"

    def get_instance_title(self):
        return f"Claim from {self.view.entry.computed_str}"

    def get_create_title(self):
        return f"New Claim for {self.view.entry.computed_str}"


class ClaimTradeTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"Claims for Customer Trade {str(self.view.trade)}"

    def get_instance_title(self):
        return f"Claim for Customer Trade {str(self.view.trade)}"

    def get_create_title(self):
        return f"New Claim for Customer Trade {str(self.view.trade)}"


class ConsolidatedTradeSummaryTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Consolidated Trade Summary"


class CumulativeNNMChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Cumulative NNM Chart"


class ConsolidatedTradeSummaryDistributionChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Consolidated Trade Summary Distribution Chart"


class ProfitAndLossPandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Profit & Loss - Averaging Method (Experimental)"


class NegativeTermimalAccountPerProductTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Negative Sub-Accounts"



import json
from datetime import datetime
from io import BytesIO
from typing import Optional

from django.db import models
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.io.backends import AbstractDataBackend, register

from ..utils import process_request
from .mixin import DataBackendMixin


@register("Fees", provider_key="ubs", save_data_in_import_source=True, passive_only=True)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    def __init__(
        self, import_credential: Optional[models.Model] = None, ubs_bank: Optional[models.Model] = None, **kwargs
    ):
        if not ubs_bank:
            raise ValueError("The ubs company objects needs to be passed to this backend")
        self.ubs_bank = ubs_bank
        if not import_credential or not import_credential.authentication_token:
            raise ValueError("UBS backend needs a valid import credential object")
        self.authentication_token = import_credential.authentication_token

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()

        mngt_fees_endpoint = "https://neo.ubs.com/api/ged-amc/external/fee/v1/management/{0}"
        perf_fees_endpoint = "https://neo.ubs.com/api/ged-amc/external/fee/v1/performance/{0}"
        if obj_external_ids:
            for external_id in obj_external_ids:
                start = kwargs.get("start", None)
                if not start:
                    start = global_preferences_registry.manager()["wbfdm__default_start_date_historical_import"]
                mngt_res = process_request(
                    self.authentication_token,
                    mngt_fees_endpoint.format(external_id),
                    {"fromDate": start.strftime("%Y-%m-%d"), "toDate": execution_date.strftime("%Y-%m-%d")},
                )
                perf_res = process_request(
                    self.authentication_token,
                    perf_fees_endpoint.format(external_id),
                    {"fromDate": start.strftime("%Y-%m-%d"), "toDate": execution_date.strftime("%Y-%m-%d")},
                )

                if mngt_res or perf_res:
                    res_json = {
                        "performance_fees": perf_res.get("fees", []),
                        "management_fees": mngt_res.get("fees", []),
                        "isin": external_id,
                    }

                    if res_json:
                        content_file = BytesIO()
                        content_file.write(json.dumps(res_json).encode())
                        file_name = f"ubs_fees_{external_id}_{start:%Y-%m-%d}_{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
                        yield file_name, content_file



from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbfdm.viewsets.configs.buttons.instruments import InstrumentButtonViewConfig


class ProductButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()

    def get_custom_instance_buttons(self):
        return {
            bt.DropDownButton(
                label="Commission",
                icon=WBIcon.UNFOLD.icon,
                buttons=(
                    bt.WidgetButton(key="claims", label="Claims"),
                    bt.WidgetButton(key="aum", label="AUM per account"),
                ),
            ),
            *InstrumentButtonViewConfig(self.view, self.request, self.instance).get_custom_instance_buttons(),
            bt.DropDownButton(
                label="Risk Management",
                icon=WBIcon.UNFOLD.icon,
                buttons=[
                    bt.WidgetButton(
                        key="risk_rules",
                        label="Rules",
                        icon=WBIcon.CONFIGURE.icon,
                    ),
                    bt.WidgetButton(
                        key="risk_incidents",
                        label="Incidents",
                        icon=WBIcon.WARNING.icon,
                    ),
                ],
            ),
        }


class ProductCustomerButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(
                key="historical-chart",
                label="Historical Chart",
                icon=WBIcon.STATS.icon,
            )
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



import json
from contextlib import suppress
from datetime import datetime

from wbportfolio.models import Product


def parse(import_source):
    data = []
    with suppress(KeyError):
        series_data = json.loads(import_source.file.read())["payload"]["series"]
        for series in series_data:
            isin = series["item"]["priceIdentifier"]
            if Product.objects.filter(isin=isin).exists():  # ensure the timeseries contain data for products we handle
                for point in series["points"]:
                    data.append(
                        {
                            "instrument": {"isin": isin},
                            "date": datetime.fromtimestamp(int(point["timestamp"]) / 1000).strftime("%Y-%m-%d"),
                            "net_value": point["close"],
                        }
                    )
    return {"data": data}



from decimal import Decimal
from typing import TYPE_CHECKING

from django.db import models
from django.db.models import Case, F, Q, When
from django.db.models.functions import Cast
from django.utils.translation import gettext_lazy as _
from wbcore.models import WBModel
from wbcrm.models import Account
from wbfdm.models.instruments import InstrumentPrice

from wbportfolio.models.transactions.claim import Claim

if TYPE_CHECKING:
    from wbportfolio.models import AccountReconciliation


class AccountReconciliationLineQuerySet(models.QuerySet):
    def update_or_create_for_reconciliation(self, reconciliation: "AccountReconciliation"):
        holding_per_product = (
            Claim.objects.filter(
                status=Claim.Status.APPROVED,
                account__in=reconciliation.account.get_descendants(include_self=True),
                account__status=Account.Status.OPEN,
                date__lte=reconciliation.reconciliation_date,
            )
            .values("product")
            .annotate(
                holdings=models.Sum("shares"),
                price=InstrumentPrice.subquery_closest_value(
                    "net_value", val_date=reconciliation.reconciliation_date, instrument_pk_name="product_id"
                ),
                price_date=InstrumentPrice.subquery_closest_value(
                    "date", val_date=reconciliation.reconciliation_date, instrument_pk_name="product_id"
                ),
            )
            .filter(~Q(holdings=0))
            .values("product_id", "holdings", "price", "price_date")
        )
        lines = []
        for holding in holding_per_product:
            line = AccountReconciliationLine(
                reconciliation=reconciliation,
                product_id=holding["product_id"],
                shares=holding["holdings"],
                shares_external=holding["holdings"],
                price=holding["price"],
                price_date=holding["price_date"],
            )
            line.calculate_fields()
            lines.append(line)

        return AccountReconciliationLine.objects.bulk_create(
            lines,
            update_conflicts=True,
            update_fields=["shares", "nominal_value", "price", "price_date"],
            unique_fields=["reconciliation", "product"],
        )

    def annotate_currency(self) -> "AccountReconciliationLineQuerySet":
        return self.annotate(currency=F("product__currency__symbol"))

    def annotate_currency_key(self) -> "AccountReconciliationLineQuerySet":
        return self.annotate(currency_key=F("product__currency__key"))

    def annotate_assets_under_management(self) -> "AccountReconciliationLineQuerySet":
        return self.annotate(assets_under_management=F("shares") * F("price"))

    def annotate_assets_under_management_external(self) -> "AccountReconciliationLineQuerySet":
        return self.annotate(assets_under_management_external=F("shares_external") * F("price"))

    def annotate_is_equal(self) -> "AccountReconciliationLineQuerySet":
        return self.annotate(
            is_equal=models.Case(
                models.When(shares=F("shares_external"), then=True), default=False, output_field=models.BooleanField()
            )
        )

    def annotate_shares_diff(self) -> "AccountReconciliationLineQuerySet":
        return self.annotate(shares_diff=F("shares_external") - F("shares"))

    def annotate_pct_diff(self) -> "AccountReconciliationLineQuerySet":
        ff = models.FloatField()
        return self.annotate(
            pct_diff=Case(
                When(~Q(shares=0), then=(Cast("shares_external", ff) - Cast("shares", ff)) / Cast("shares", ff)),
                default=None,
            )
        )

    def annotate_nominal_value_diff(self) -> "AccountReconciliationLineQuerySet":
        return self.annotate(nominal_value_diff=F("nominal_value_external") - F("nominal_value"))

    def annotate_assets_under_management_diff(self) -> "AccountReconciliationLineQuerySet":
        """Annotates the AuM diff. This will fail if annotate_assets_under_management was not called beforehand"""
        return self.annotate(
            assets_under_management_diff=F("assets_under_management_external") - F("assets_under_management")
        )


class AccountReconciliationLine(WBModel):
    reconciliation = models.ForeignKey(
        to="wbportfolio.AccountReconciliation",
        related_name="lines",
        on_delete=models.CASCADE,
    )

    product = models.ForeignKey(
        to="wbportfolio.Product",
        related_name="account_reconciliation_lines",
        on_delete=models.CASCADE,
    )

    price = models.DecimalField(
        decimal_places=4,
        max_digits=18,
        default=Decimal(0),
        help_text=_("The last share price of the product"),
    )

    price_date = models.DateField()

    shares = models.DecimalField(
        decimal_places=4,
        max_digits=18,
        default=Decimal(0),
        help_text=_("The number of shares computed through the Workbench"),
    )
    nominal_value = models.DecimalField(
        decimal_places=4,
        max_digits=18,
        default=Decimal(0),
        help_text=_("The nominal value computed through the Workbench"),
    )
    shares_external = models.DecimalField(
        decimal_places=4,
        max_digits=18,
        default=Decimal(0),
        help_text=_(
            "The number of shares externally provided through the Account holder. Initially set to the number of shares computed through the Workbench"
        ),
    )
    nominal_value_external = models.DecimalField(
        decimal_places=4,
        max_digits=18,
        default=Decimal(0),
        help_text=_(
            "The nominal value externally provided through the Account holder. Initially set to the number of shares computed through the Workbench"
        ),
    )

    objects = AccountReconciliationLineQuerySet.as_manager()

    class Meta:
        verbose_name = _("Account Reconciliation (Product)")
        verbose_name_plural = _("Account Reconciliation (Products)")
        constraints = [
            models.UniqueConstraint(fields=["reconciliation", "product"], name="unique_reconcilation_product"),
        ]

    def save(self, *args, **kwargs):
        self.calculate_fields()
        super().save(*args, **kwargs)

    def calculate_fields(self):
        if self.shares is not None:
            self.nominal_value = self.shares * self.product.share_price
        elif self.nominal_value is not None:
            self.shares = self.nominal_value / self.product.share_price

        if self.shares_external is not None:
            self.nominal_value_external = self.shares_external * self.product.share_price
        elif self.nominal_value_external is not None:
            self.shares_external = self.nominal_value_external / self.product.share_price

    def __str__(self) -> str:
        return f"{self.reconciliation}: {self.product} {self.shares} shares"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:accountreconciliationline"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:accountreconciliationlinerepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{reconciliation}}"



from contextlib import suppress
from datetime import date, datetime

import pandas as pd
import plotly.graph_objects as go
from django.db.models import Exists, F, OuterRef, Q, Sum
from django.utils.dateparse import parse_date
from django.utils.functional import cached_property
from plotly.subplots import make_subplots
from wbcore import viewsets
from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.filters import DjangoFilterBackend
from wbcore.pandas import fields as pf
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.serializers import decorator
from wbcore.utils.date import get_date_interval_from_request
from wbcore.utils.figures import (
    get_default_timeserie_figure,
    get_hovertemplate_timeserie,
)
from wbcore.utils.strings import format_number
from wbfdm.contrib.metric.viewsets.mixins import InstrumentMetricMixin
from wbfdm.models import Instrument

from wbportfolio.filters import (
    AssetPositionFilter,
    AssetPositionInstrumentFilter,
    AssetPositionPortfolioFilter,
    AssetPositionUnderlyingInstrumentChartFilter,
    CashPositionPortfolioFilterSet,
    CompositionContributionChartFilter,
    CompositionModelPortfolioPandasFilter,
    ContributionChartFilter,
)
from wbportfolio.import_export.resources.assets import AssetPositionResource
from wbportfolio.metric.backends.portfolio_base import (
    PORTFOLIO_CAPITAL_EMPLOYED,
    PORTFOLIO_EBIT,
    PORTFOLIO_LIABILITIES,
    PORTFOLIO_ROCE,
    PORTFOLIO_TOTAL_ASSETS,
)
from wbportfolio.metric.backends.portfolio_esg import PORTFOLIO_ESG_KEYS
from wbportfolio.models import (
    AssetPosition,
    InstrumentPortfolioThroughModel,
    Portfolio,
    Product,
)
from wbportfolio.serializers.assets import (
    AssetPositionAggregatedPortfolioModelSerializer,
    AssetPositionInstrumentModelSerializer,
    AssetPositionModelSerializer,
    AssetPositionPortfolioModelSerializer,
    CashPositionPortfolioModelSerializer,
)

from .configs import (
    AssetPositionButtonConfig,
    AssetPositionDisplayConfig,
    AssetPositionEndpointConfig,
    AssetPositionInstrumentButtonConfig,
    AssetPositionInstrumentDisplayConfig,
    AssetPositionInstrumentEndpointConfig,
    AssetPositionInstrumentTitleConfig,
    AssetPositionPortfolioButtonConfig,
    AssetPositionPortfolioDisplayConfig,
    AssetPositionPortfolioEndpointConfig,
    AssetPositionPortfolioTitleConfig,
    AssetPositionTitleConfig,
    AssetPositionUnderlyingInstrumentChartEndpointConfig,
    AssetPositionUnderlyingInstrumentChartTitleConfig,
    CashPositionPortfolioDisplayConfig,
    CashPositionPortfolioEndpointConfig,
    CashPositionPortfolioTitleConfig,
    CompositionModelPortfolioPandasDisplayConfig,
    CompositionModelPortfolioPandasEndpointConfig,
    CompositionModelPortfolioPandasTitleConfig,
    ContributorPortfolioChartEndpointConfig,
    ContributorPortfolioChartTitleConfig,
)
from .mixins import UserPortfolioRequestPermissionMixin


class AssetPositionModelViewSet(
    UserPortfolioRequestPermissionMixin,
    InternalUserPermissionMixin,
    viewsets.ReadOnlyModelViewSet,
):
    IDENTIFIER = "wbportfolio:assetposition"
    IMPORT_ALLOWED = False

    queryset = AssetPosition.objects.all()
    serializer_class = AssetPositionModelSerializer
    filterset_class = AssetPositionFilter

    ordering_fields = [
        "total_value_fx_portfolio",
        "total_value_fx_usd",
        "total_value",
        "portfolio__name",
        "portfolio_created__name",
        "underlying_quote",
        "underlying_quote_name",
        "underlying_quote_ticker",
        "underlying_quote_isin",
        "price",
        "currency__key",
        "currency_fx_rate",
        "shares",
        "date",
        "asset_valuation_date",
        "weighting",
        "market_share",
        "liquidity",
    ]
    ordering = ["-weighting"]
    search_fields = ["underlying_quote_ticker", "underlying_quote_name", "underlying_quote_isin"]

    display_config_class = AssetPositionDisplayConfig
    button_config_class = AssetPositionButtonConfig
    endpoint_config_class = AssetPositionEndpointConfig
    title_config_class = AssetPositionTitleConfig

    def get_resource_class(self):
        return AssetPositionResource

    def get_aggregates(self, queryset, paginated_queryset):
        aggregates = super().get_aggregates(queryset, paginated_queryset)
        if queryset.exists():
            total_value_fx_usd = queryset.aggregate(s=Sum(F("total_value_fx_usd")))["s"]
            weighting = queryset.aggregate(s=Sum(F("weighting")))["s"]
            aggregates.update(
                {
                    "weighting": {"Σ": format_number(weighting, decimal=4)},
                    "total_value_fx_usd": {"Σ": format_number(total_value_fx_usd)},
                }
            )
        return aggregates

    def get_queryset(self):
        if self.is_analyst:
            return (
                super()
                .get_queryset()
                .annotate(
                    underlying_quote_isin=F("underlying_quote__isin"),
                    underlying_quote_ticker=F("underlying_quote__ticker"),
                    underlying_quote_name=F("underlying_quote__name"),
                )
                .select_related(
                    "underlying_quote",
                    "currency",
                    "portfolio",
                    "exchange",
                    "portfolio_created",
                )
            )
        return AssetPosition.objects.none()


# Portfolio Viewsets


class AssetPositionPortfolioModelViewSet(InstrumentMetricMixin, AssetPositionModelViewSet):
    METRIC_KEYS = (
        PORTFOLIO_EBIT,
        PORTFOLIO_TOTAL_ASSETS,
        PORTFOLIO_LIABILITIES,
        PORTFOLIO_CAPITAL_EMPLOYED,
        PORTFOLIO_ROCE,
        *PORTFOLIO_ESG_KEYS,
    )
    METRIC_BASKET_LABEL = "portfolio"
    METRIC_INSTRUMENT_LABEL = "underlying_quote"
    METRIC_SHOW_BY_DEFAULT = False
    METRIC_SHOW_FILTERS = True

    @property
    def metric_date(self):
        if date_str := self.request.GET.get("date"):
            return parse_date(date_str)

    @property
    def metric_basket(self):
        return self.portfolio

    filterset_class = AssetPositionPortfolioFilter
    display_config_class = AssetPositionPortfolioDisplayConfig
    button_config_class = AssetPositionPortfolioButtonConfig
    title_config_class = AssetPositionPortfolioTitleConfig
    endpoint_config_class = AssetPositionPortfolioEndpointConfig

    def get_serializer_class(self):
        if self.request.GET.get("aggregate", "false") == "true":
            return AssetPositionAggregatedPortfolioModelSerializer
        return AssetPositionPortfolioModelSerializer

    def get_aggregates(self, queryset, paginated_queryset):
        if not queryset.exists():
            return {}
        weighting = queryset.aggregate(s=Sum(F("weighting")))["s"]
        total_value_fx_portfolio = queryset.aggregate(s=Sum(F("total_value_fx_portfolio")))["s"]
        aggregates = super().get_aggregates(queryset, paginated_queryset)
        aggregates["total_value_fx_portfolio"] = {"Σ": format_number(total_value_fx_portfolio)}
        aggregates["weighting"] = {"Σ": format_number(weighting, decimal=4)}
        return aggregates

    def get_queryset(self):
        if self.has_portfolio_access:
            return (
                super()
                .get_queryset()
                .filter(portfolio=self.portfolio)
                .select_related("underlying_quote", "currency", "exchange", "portfolio_created")
            )
        return AssetPosition.objects.none()


# Underlying Assets viewsets
class AssetPositionInstrumentModelViewSet(AssetPositionModelViewSet):
    filterset_class = AssetPositionInstrumentFilter
    serializer_class = AssetPositionInstrumentModelSerializer
    display_config_class = AssetPositionInstrumentDisplayConfig
    title_config_class = AssetPositionInstrumentTitleConfig
    button_config_class = AssetPositionInstrumentButtonConfig
    endpoint_config_class = AssetPositionInstrumentEndpointConfig

    def get_aggregates(self, queryset, paginated_queryset):
        queryset = queryset.filter(is_invested=True)
        if queryset.exists():
            total_value_fx_usd = queryset.aggregate(s=Sum(F("total_value_fx_usd")))["s"]
            total_shares = queryset.aggregate(s=Sum(F("shares")))["s"]
            total_market_share = queryset.aggregate(s=Sum(F("market_share")))["s"]
            return {
                "shares": {"Σ": format_number(total_shares)},
                "total_value_fx_usd": {"Σ": format_number(total_value_fx_usd)},
                "market_share": {"Σ": format_number(total_market_share)},
            }
        return {}

    def get_queryset(self):
        qs = super().get_queryset().filter(underlying_quote__in=self.instrument.get_descendants(include_self=True))
        if self.request.GET.get("filter_last_positions", "false") == "true":
            if self.instrument.assets.exists():
                qs = qs.filter(date=self.instrument.assets.latest("date").date)
        return qs


class CashPositionPortfolioPandasAPIView(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet
):
    queryset = AssetPosition.objects.all()
    display_config_class = CashPositionPortfolioDisplayConfig
    title_config_class = CashPositionPortfolioTitleConfig
    endpoint_config_class = CashPositionPortfolioEndpointConfig

    serializer_class = CashPositionPortfolioModelSerializer
    filterset_class = CashPositionPortfolioFilterSet

    search_fields = ["portfolio_name"]
    ordering_fields = ["portfolio_name", "total_value_fx_usd", "portfolio_weight"]

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="portfolio", label="ID"),
            pf.CharField(key="portfolio_name", label="Portfolio"),
            pf.FloatField(
                key="total_value_fx_usd",
                label="Total Value",
                precision=2,
                decorators=(decorator(decorator_type="text", position="right", value="$"),),
            ),
            pf.FloatField(key="portfolio_weight", label="Total portfolio value", precision=2, percent=True),
        )
    )

    def get_aggregates(self, request, df):
        if not df.empty:
            sum_total_value_fx_usd = df.total_value_fx_usd.sum()
            return {
                "total_value_fx_usd": {"Σ": format_number(sum_total_value_fx_usd)},
            }
        return {}

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame()
        if queryset.exists():
            df = pd.DataFrame(
                queryset.values(
                    "underlying_instrument__is_cash",
                    "total_value_fx_portfolio",
                    "fx_rate",
                    "portfolio__name",
                    "portfolio",
                )
            )
            df_name = (
                df[["portfolio", "portfolio__name"]]
                .groupby("portfolio")
                .agg("first")
                .rename(columns={"portfolio__name": "portfolio_name"})
            )

            df["total_value_fx_usd"] = df.total_value_fx_portfolio * df.fx_rate
            df_total = (
                df[["portfolio", "total_value_fx_usd"]]
                .groupby("portfolio")
                .sum()
                .rename(columns={"total_value_fx_usd": "total_portfolio_fx_usd"})
            )
            df = df[df["underlying_instrument__is_cash"]]
            df = df[["portfolio", "total_value_fx_usd"]].groupby("portfolio").sum()
            if not df.empty:
                df = df[df["total_value_fx_usd"] != 0]
                df = pd.concat([df, df_total, df_name], axis=1).dropna(how="any")
                df["portfolio_weight"] = df.total_value_fx_usd / df.total_portfolio_fx_usd
                return df.reset_index()
        return df

    def get_queryset(self):
        if self.is_portfolio_manager:
            if date_str := self.request.GET.get("date", None):
                val_date = datetime.strptime(date_str, "%Y-%m-%d").date()
            else:
                val_date = date.today()
            active_products = Product.active_objects.filter_active_at_date(val_date)
            return (
                super()
                .get_queryset()
                .annotate(
                    fx_rate=CurrencyFXRates.get_fx_rates_subquery(
                        "date", currency="portfolio__currency", lookup_expr="exact"
                    ),
                    is_product_portfolio=Exists(
                        InstrumentPortfolioThroughModel.objects.filter(
                            portfolio=OuterRef("portfolio"), instrument__in=active_products
                        )
                    ),
                )
                .filter(is_product_portfolio=True)
            )
        return AssetPosition.objects.none()


# ##### CHART VIEWS #####


class ContributorPortfolioChartView(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
    filterset_class = ContributionChartFilter
    filter_backends = (DjangoFilterBackend,)
    IDENTIFIER = "wbportfolio:portfolio-contributor"
    queryset = AssetPosition.objects.all()

    title_config_class = ContributorPortfolioChartTitleConfig
    endpoint_config_class = ContributorPortfolioChartEndpointConfig

    ROW_HEIGHT: int = 20

    @property
    def min_height(self):
        if hasattr(self, "nb_rows"):
            return self.nb_rows * self.ROW_HEIGHT
        return "300px"

    @cached_property
    def hedged_currency(self) -> Currency | None:
        if "hedged_currency" in self.request.GET:
            with suppress(Currency.DoesNotExist):
                return Currency.objects.get(pk=self.request.GET["hedged_currency"])

    @cached_property
    def show_lookthrough(self) -> bool:
        return self.portfolio.is_composition and self.request.GET.get("show_lookthrough", "false").lower() == "true"

    def get_filterset_class(self, request):
        if self.portfolio.is_composition:
            return CompositionContributionChartFilter
        return ContributionChartFilter

    def get_plotly(self, queryset):
        fig = go.Figure()
        data = []
        if self.show_lookthrough:
            d1, d2 = get_date_interval_from_request(self.request)
            for _d in pd.date_range(d1, d2):
                for pos in self.portfolio.get_lookthrough_positions(_d.date()):
                    data.append(
                        [
                            pos.date,
                            pos.initial_price,
                            pos.initial_currency_fx_rate,
                            pos.underlying_instrument_id,
                            pos.weighting,
                        ]
                    )
        else:
            data = queryset.annotate_hedged_currency_fx_rate(self.hedged_currency).values_list(
                "date", "price", "hedged_currency_fx_rate", "underlying_instrument", "weighting"
            )
        df = Portfolio.get_contribution_df(data).rename(columns={"group_key": "underlying_instrument"})
        if not df.empty:
            df = df[["contribution_total", "contribution_forex", "underlying_instrument"]].sort_values(
                by="contribution_total", ascending=True
            )

            df["instrument_id"] = df.underlying_instrument.map(
                dict(Instrument.objects.filter(id__in=df["underlying_instrument"]).values_list("id", "name_repr"))
            )
            df_forex = df[["instrument_id", "contribution_forex"]]
            df_forex = df_forex[df_forex.contribution_forex != 0]

            contribution_equity = df.contribution_total - df.contribution_forex

            text_forex = df_forex.contribution_forex.apply(lambda x: f"{x:,.2%}")
            text_equity = contribution_equity.apply(lambda x: f"{x:,.2%}")
            setattr(self, "nb_rows", df.shape[0])
            fig.add_trace(
                go.Bar(
                    y=df.instrument_id,
                    x=contribution_equity,
                    name="Contribution Equity",
                    orientation="h",
                    marker=dict(
                        color="rgba(247,110,91,0.6)",
                        line=dict(color="rgb(247,110,91,1.0)", width=2),
                    ),
                    text=text_equity.values,
                    textposition="auto",
                )
            )
            fig.add_trace(
                go.Bar(
                    y=df_forex.instrument_id,
                    x=df_forex.contribution_forex,
                    name="Contribution Forex",
                    orientation="h",
                    marker=dict(
                        color="rgba(58, 71, 80, 0.6)",
                        line=dict(color="rgba(58, 71, 80, 1.0)", width=2),
                    ),
                    text=text_forex.values,
                    textposition="outside",
                )
            )
            fig.update_layout(
                barmode="relative",
                xaxis=dict(showgrid=False, showline=False, zeroline=False, tickformat=".2%"),
                yaxis=dict(showgrid=False, showline=False, zeroline=False, tickmode="linear"),
                margin=dict(b=0, r=20, l=20, t=0, pad=20),
                paper_bgcolor="rgba(0,0,0,0)",
                plot_bgcolor="rgba(0,0,0,0)",
                font=dict(family="roboto", size=12, color="black"),
                bargap=0.3,
            )
            # fig = get_horizontal_barplot(df, x_label="contribution_total", y_label="name")
        return fig

    def parse_figure_dict(self, figure_dict: dict[str, any]) -> dict[str, any]:
        figure_dict = super().parse_figure_dict(figure_dict)
        figure_dict["style"]["minHeight"] = self.min_height
        return figure_dict

    def get_queryset(self):
        if self.has_portfolio_access:
            return super().get_queryset().filter(portfolio=self.portfolio)
        return AssetPosition.objects.none()


class AssetPositionUnderlyingInstrumentChartViewSet(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
    IDENTIFIER = "wbportfolio:assetpositionchart"

    queryset = AssetPosition.objects.all()

    title_config_class = AssetPositionUnderlyingInstrumentChartTitleConfig
    endpoint_config_class = AssetPositionUnderlyingInstrumentChartEndpointConfig
    filterset_class = AssetPositionUnderlyingInstrumentChartFilter

    def get_queryset(self):
        return AssetPosition.objects.filter(underlying_quote__in=self.instrument.get_descendants(include_self=True))

    def get_plotly(self, queryset):
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        fig = get_default_timeserie_figure(fig)
        if queryset.exists():
            df_price = (pd.DataFrame(queryset.values("date", "price_fx_usd"))).groupby("date").first()

            df_price = df_price.where(pd.notnull(df_price), 0)
            fig.add_trace(
                go.Scatter(
                    x=df_price.index, y=df_price.price_fx_usd, mode="lines", marker_color="green", name="Price"
                ),
                secondary_y=False,
            )

            df_weight = pd.DataFrame(queryset.values("date", "weighting", "portfolio__name"))
            df_weight = df_weight.where(pd.notnull(df_weight), 0)
            df_weight = df_weight.groupby(["date", "portfolio__name"]).sum().reset_index()
            for portfolio_name, df_tmp in df_weight.groupby("portfolio__name"):
                fig.add_trace(
                    go.Scatter(
                        x=df_tmp.date,
                        y=df_tmp.weighting,
                        hovertemplate=get_hovertemplate_timeserie(is_percent=True),
                        mode="lines",
                        name=f"Allocation: {portfolio_name}",
                    ),
                    secondary_y=True,
                )

            # Set x-axis title
            fig.update_xaxes(title_text="Date")
            # Set y-axes titles
            fig.update_yaxes(
                title_text="<b>Price</b>",
                secondary_y=False,
                titlefont=dict(color="green"),
                tickfont=dict(color="green"),
            )
            fig.update_yaxes(
                title_text="<b>Portfolio Allocation (%)</b>",
                secondary_y=True,
                titlefont=dict(color="blue"),
                tickfont=dict(color="blue"),
            )

        return fig


class CompositionModelPortfolioPandasView(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet
):
    IDENTIFIER = "wbportfolio:topdowncomposition"
    filterset_class = CompositionModelPortfolioPandasFilter
    queryset = AssetPosition.objects.all()

    display_config_class = CompositionModelPortfolioPandasDisplayConfig
    title_config_class = CompositionModelPortfolioPandasTitleConfig
    endpoint_config_class = CompositionModelPortfolioPandasEndpointConfig

    @cached_property
    def val_date(self) -> date:
        return parse_date(self.request.GET["date"])

    @cached_property
    def dependant_portfolios(self):
        return Portfolio.objects.filter(
            Q(depends_on__in=[self.portfolio]) | Q(dependent_portfolios__in=[self.portfolio])
        )

    def get_queryset(self):
        if self.has_portfolio_access:
            return super().get_queryset().filter(portfolio=self.portfolio)
        return AssetPosition.objects.none()

    def get_pandas_fields(self, request):
        fields = [
            pf.PKField(key="underlying_instrument", label="ID"),
            pf.CharField(key="underlying_instrument_repr", label="Instrument"),
        ]
        if not self.portfolio.only_weighting:
            fields.append(pf.FloatField(key=f"shares_{self.portfolio.id}", label=str(self.portfolio)))
        fields.append(pf.FloatField(key=f"weighting_{self.portfolio.id}", label=str(self.portfolio), percent=True))

        for portfolio in self.dependant_portfolios:
            if not portfolio.only_weighting:
                fields.append(pf.FloatField(key=f"shares_{portfolio.id}", label=str(portfolio)))
                fields.append(pf.FloatField(key=f"difference_{portfolio.id}", label=str(portfolio)))
            else:
                fields.append(pf.FloatField(key=f"weighting_{portfolio.id}", label=str(portfolio), percent=True))
                fields.append(pf.FloatField(key=f"difference_{portfolio.id}", label=str(portfolio), percent=True))

        return pf.PandasFields(fields=fields)

    search_fields = ["underlying_instrument_repr"]
    ordering_fields = ["underlying_instrument_repr", "model_portfolio"]

    def get_dataframe(self, request, queryset, **kwargs):
        rows = []
        for portfolio in [*self.dependant_portfolios, self.portfolio]:
            rows.extend(
                list(
                    map(
                        lambda x: {
                            "underlying_instrument": x.underlying_instrument.id,
                            "weighting": x.weighting,
                            "shares": x.shares,
                            "portfolio": portfolio.id,
                        },
                        portfolio.get_positions(self.val_date),
                    )
                )
            )
        df = pd.DataFrame(rows, columns=["underlying_instrument", "weighting", "shares", "portfolio"])
        df = df.pivot_table(
            index="underlying_instrument",
            columns=["portfolio"],
            values=["weighting", "shares"],
            aggfunc="sum",
            fill_value=0,
        )
        for portfolio in self.dependant_portfolios:
            with suppress(KeyError):
                if portfolio.only_weighting:
                    df["difference", portfolio.id] = df["weighting"][self.portfolio.id] - df["weighting"][portfolio.id]
                else:
                    df["difference", portfolio.id] = df["shares"][self.portfolio.id] - df["shares"][portfolio.id]

        df.columns = ["_".join([str(item) for item in col]) for col in df.columns.to_flat_index()]
        df = df.reset_index()

        return df

    def manipulate_dataframe(self, df):
        df["underlying_instrument_repr"] = df["underlying_instrument"].map(
            dict(Instrument.objects.filter(id__in=df["underlying_instrument"]).values_list("id", "computed_str"))
        )
        return df



from typing import Optional

from django.shortcuts import get_object_or_404
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbportfolio.models import Trade, TradeProposal

NEGATIVE_RED_FORMATTING = [
    dp.FormattingRule(
        style={
            "color": WBColor.RED_DARK.value,
            "fontWeight": "bold",
        },
        condition=("<", 0),
    )
]

SHARE_FORMATTING = dp.Formatting(
    column="shares",
    formatting_rules=[
        dp.FormattingRule(
            style={
                "color": WBColor.RED_DARK.value,
                # "fontWeight": "bold",
            },
            condition=("<", 0),
        )
    ],
)

TRADE_STATUS_LEGENDS = dp.Legend(
    key="status",
    items=[
        dp.LegendItem(
            icon=WBColor.RED_LIGHT.value,
            label=Trade.Status.FAILED.label,
            value=Trade.Status.FAILED.value,
        ),
        dp.LegendItem(
            icon=WBColor.BLUE_LIGHT.value,
            label=Trade.Status.DRAFT.label,
            value=Trade.Status.DRAFT.value,
        ),
        dp.LegendItem(
            icon=WBColor.YELLOW_LIGHT.value,
            label=Trade.Status.SUBMIT.label,
            value=Trade.Status.SUBMIT.value,
        ),
        dp.LegendItem(
            icon=WBColor.GREEN_LIGHT.value,
            label=Trade.Status.EXECUTED.label,
            value=Trade.Status.EXECUTED.value,
        ),
        dp.LegendItem(
            icon=WBColor.GREEN.value,
            label=Trade.Status.CONFIRMED.label,
            value=Trade.Status.CONFIRMED.value,
        ),
    ],
)

TRADE_STATUS_FORMATTING = dp.Formatting(
    column="status",
    formatting_rules=[
        dp.FormattingRule(
            style={"backgroundColor": WBColor.RED_LIGHT.value},
            condition=("==", Trade.Status.FAILED.value),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
            condition=("==", Trade.Status.DRAFT.value),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
            condition=("==", Trade.Status.SUBMIT.value),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
            condition=("==", Trade.Status.EXECUTED.value),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.GREEN.value},
            condition=("==", Trade.Status.CONFIRMED.value),
        ),
    ],
)


class TradeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="transaction_subtype", label="Type", width=Unit.PIXEL(100)),
                dp.Field(key="transaction_date", label="Trade Date", width=Unit.PIXEL(150)),
                dp.Field(key="underlying_instrument", label="Instrument", width=Unit.PIXEL(250)),
                dp.Field(
                    label="Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="shares", label="Shares", width=Unit.PIXEL(100)),
                        dp.Field(key="price", label="Price", width=Unit.PIXEL(100)),
                        dp.Field(key="total_value", label="Value", width=Unit.PIXEL(125)),
                        dp.Field(key="total_value_usd", label="Value ($)", width=Unit.PIXEL(125), show="open"),
                    ],
                ),
                dp.Field(
                    label="Bank",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="bank", label="Counterparty", width=Unit.PIXEL(150)),
                        dp.Field(key="custodian", label="Custodian", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="register", label="Register", width=Unit.PIXEL(200), show="open"),
                    ],
                ),
                dp.Field(
                    label="Information",
                    open_by_default=False,
                    key=None,
                    children=list(
                        filter(
                            lambda x: not (x.key == "portfolio" and "portfolio_id" in self.view.kwargs),
                            [
                                dp.Field(key="currency", label="Currency", width=Unit.PIXEL(100)),
                                dp.Field(key="portfolio", label="Portfolio", width=Unit.PIXEL(250), show="open"),
                                dp.Field(
                                    key="marked_for_deletion",
                                    label="Marked for Deletion",
                                    width=Unit.PIXEL(150),
                                    show="open",
                                ),
                                dp.Field(key="pending", label="Pending", width=Unit.PIXEL(150), show="open"),
                            ],
                        )
                    ),
                ),
            ],
            legends=[TRADE_STATUS_LEGENDS],
            formatting=[SHARE_FORMATTING, TRADE_STATUS_FORMATTING],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["transaction_date", "value_date", "."],
                [repeat_field(3, "underlying_instrument")],
                ["shares", "price", "bank"],
                ["external_id", "marked_for_deletion", "register"],
                [repeat_field(3, "comment")],
            ]
        )


class SubscriptionRedemptionDisplayConfig(TradeDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="transaction_subtype", label="Type", width=Unit.PIXEL(100)),
                dp.Field(key="transaction_date", label="Trade Date", width=Unit.PIXEL(150)),
                dp.Field(key="underlying_instrument", label="Product", width=Unit.PIXEL(250)),
                dp.Field(
                    label="Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="shares", label="Shares", width=Unit.PIXEL(100)),
                        dp.Field(key="price", label="Price", width=Unit.PIXEL(100)),
                        dp.Field(key="total_value", label="Value", width=Unit.PIXEL(125)),
                        dp.Field(key="total_value_usd", label="Value ($)", width=Unit.PIXEL(125), show="open"),
                    ],
                ),
                dp.Field(
                    label="Customer",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="claimed_shares", label="Claimed Shares", width=Unit.PIXEL(150)),
                        dp.Field(key="claims", label="Claims", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="comment", label="Comment", width=Unit.PIXEL(200)),
                    ],
                ),
                dp.Field(
                    label="Bank",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="bank", label="Bank", width=Unit.PIXEL(150)),
                        dp.Field(key="custodian", label="Custodian", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="register", label="Register", width=Unit.PIXEL(200), show="open"),
                    ],
                ),
                dp.Field(
                    label="Information",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="currency", label="Currency", width=Unit.PIXEL(100), show="open"),
                        dp.Field(key="portfolio", label="Portfolio", width=Unit.PIXEL(250), show="open"),
                        dp.Field(key="marked_for_deletion", label="Marked for Deletion", width=Unit.PIXEL(150)),
                        dp.Field(key="pending", label="Pending", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="marked_as_internal", label="Internal", width=Unit.PIXEL(150), show="open"),
                        dp.Field(key="internal_trade", label="Internal Trade", width=Unit.PIXEL(150), show="open"),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    key="completely_claimed",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label="Completely Claimed",
                            value=True,
                        )
                    ],
                ),
                dp.Legend(
                    key="completely_claimed_if_approved",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label="Completely Claimed if approved",
                            value=True,
                        )
                    ],
                ),
                dp.Legend(
                    key="pending",
                    items=[
                        dp.LegendItem(icon=WBIcon.FOLDERS_ADD.icon, label="Pending", value=True),
                    ],
                ),
                dp.Legend(
                    key="marked_for_deletion",
                    items=[
                        dp.LegendItem(icon=WBIcon.DELETE.icon, label="To Be Deleted", value=True),
                    ],
                ),
            ],
            formatting=[
                SHARE_FORMATTING,
                dp.Formatting(
                    column="completely_claimed",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", True),
                        )
                    ],
                ),
                dp.Formatting(
                    column="completely_claimed_if_approved",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", True),
                        )
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["transaction_date", "value_date", "."],
                [repeat_field(3, "underlying_instrument")],
                ["shares", "price", "bank"],
                ["external_id", "marked_for_deletion", "register"],
                [repeat_field(3, "comment")],
                ["marked_as_internal", repeat_field(2, "internal_trade")],
            ]
        )


class TradePortfolioDisplayConfig(TradeDisplayConfig):
    pass


class TradeTradeProposalDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        trade_proposal = get_object_or_404(TradeProposal, pk=self.view.kwargs.get("trade_proposal_id", None))
        fields = [
            dp.Field(
                label="Instrument",
                open_by_default=True,
                key=None,
                children=[
                    dp.Field(key="underlying_instrument", label="Name", width=Unit.PIXEL(250)),
                    dp.Field(key="underlying_instrument_isin", label="ISIN", width=Unit.PIXEL(125)),
                    dp.Field(key="underlying_instrument_ticker", label="Ticker", width=Unit.PIXEL(100)),
                    dp.Field(
                        key="underlying_instrument_refinitiv_identifier_code", label="RIC", width=Unit.PIXEL(100)
                    ),
                    dp.Field(key="underlying_instrument_instrument_type", label="Asset Class", width=Unit.PIXEL(125)),
                ],
            ),
            dp.Field(
                label="Weight",
                open_by_default=False,
                key=None,
                children=[
                    dp.Field(key="effective_weight", label="Effective Weight", show="open", width=Unit.PIXEL(150)),
                    dp.Field(key="target_weight", label="Target Weight", show="open", width=Unit.PIXEL(150)),
                    dp.Field(
                        key="weighting",
                        label="Delta Weight",
                        formatting_rules=[
                            dp.FormattingRule(
                                style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
                                condition=("<", 0),
                            ),
                            dp.FormattingRule(
                                style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
                                condition=(">", 0),
                            ),
                        ],
                        width=Unit.PIXEL(150),
                    ),
                ],
            ),
        ]
        if not trade_proposal.portfolio.only_weighting:
            fields.append(
                dp.Field(
                    label="Shares",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="effective_shares", label="Effective Shares", show="open", width=Unit.PIXEL(150)),
                        dp.Field(key="target_shares", label="Target Shares", show="open", width=Unit.PIXEL(150)),
                        dp.Field(
                            key="shares",
                            label="Shares",
                            formatting_rules=[
                                dp.FormattingRule(
                                    style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
                                    condition=("<", 0),
                                ),
                                dp.FormattingRule(
                                    style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
                                    condition=(">", 0),
                                ),
                            ],
                            width=Unit.PIXEL(150),
                        ),
                    ],
                )
            )
            fields.append(
                dp.Field(
                    label="Total Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(
                            key="effective_total_value_fx_portfolio",
                            label="Effective Total Value",
                            show="open",
                            width=Unit.PIXEL(150),
                        ),
                        dp.Field(
                            key="target_total_value_fx_portfolio",
                            label="Target Total Value",
                            show="open",
                            width=Unit.PIXEL(150),
                        ),
                        dp.Field(
                            key="total_value_fx_portfolio",
                            label="Total Value",
                            formatting_rules=[
                                dp.FormattingRule(
                                    style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
                                    condition=("<", 0),
                                ),
                                dp.FormattingRule(
                                    style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
                                    condition=(">", 0),
                                ),
                            ],
                            width=Unit.PIXEL(150),
                        ),
                    ],
                )
            )
        fields.append(
            dp.Field(
                label="Information",
                open_by_default=False,
                key=None,
                children=[
                    dp.Field(
                        key="transaction_subtype",
                        label="Direction",
                        formatting_rules=[
                            dp.FormattingRule(
                                style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
                                condition=("==", Trade.Type.SELL.name),
                            ),
                            dp.FormattingRule(
                                style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
                                condition=("==", Trade.Type.DECREASE.name),
                            ),
                            dp.FormattingRule(
                                style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
                                condition=("==", Trade.Type.INCREASE.name),
                            ),
                            dp.FormattingRule(
                                style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
                                condition=("==", Trade.Type.BUY.name),
                            ),
                            dp.FormattingRule(
                                style={"color": WBColor.GREY.value, "fontWeight": "bold"},
                                condition=("==", Trade.Type.NO_CHANGE.name),
                            ),
                        ],
                        width=Unit.PIXEL(125),
                    ),
                    dp.Field(key="comment", label="Comment", width=Unit.PIXEL(250)),
                    dp.Field(key="order", label="Order", show="open", width=Unit.PIXEL(100)),
                ],
            )
        )
        return dp.ListDisplay(
            fields=fields,
            legends=[TRADE_STATUS_LEGENDS],
            formatting=[TRADE_STATUS_FORMATTING],
        )

    def get_instance_display(self) -> Display:
        trade_proposal = get_object_or_404(TradeProposal, pk=self.view.kwargs.get("trade_proposal_id", None))

        fields = [
            ["company", "security", "underlying_instrument"],
            ["effective_weight", "target_weight", "weighting"],
        ]
        if not trade_proposal.portfolio.only_weighting:
            fields.append(["effective_shares", "target_shares", "shares"])
        fields.append([repeat_field(3, "comment")])
        return create_simple_display(fields)



from .claim import Claim
from .dividends import DividendTransaction
from .fees import FeeCalculation, Fees
from .trade_proposals import TradeProposal
from .rebalancing import RebalancingModel, Rebalancer
from .trades import Trade



from wbfdm.models.instruments import InstrumentType

from .mixins.instruments import PMSInstrumentAbstractModel


class Index(PMSInstrumentAbstractModel):
    def pre_save(self):
        super().pre_save()
        self.instrument_type = InstrumentType.INDEX
        if "market_data" not in self.dl_parameters:
            # we default to the internal dataloader
            self.dl_parameters["market_data"] = {
                "path": "wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader"
            }
        self.is_managed = True

    class Meta:
        verbose_name = "Index"
        verbose_name_plural = "Indexes"



# Generated by Django 4.2.9 on 2024-01-30 16:05

from django.db import migrations, models
from tqdm import tqdm


def migrate_fees_date(apps, schema_editor):
    Fees = apps.get_model("wbportfolio", "Fees")
    fees = []

    for fee in tqdm(Fees.objects.all(), total=Fees.objects.count()):
        fee.fee_date = fee.transaction_date
        fees.append(fee)
        if len(fees) % 10000:
            Fees.objects.bulk_update(fees, fields=["fee_date"])
            fees = []
    Fees.objects.bulk_update(fees, fields=["fee_date"])


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0049_trade_claimed_shares"),
    ]

    operations = [
        migrations.AddField(
            model_name="fees",
            name="fee_date",
            field=models.DateField(null=True, blank=True),
        ),
        migrations.RunPython(migrate_fees_date),
        migrations.AlterField(
            model_name="fees",
            name="fee_date",
            field=models.DateField(),
        ),
        migrations.AddIndex(
            model_name="fees",
            index=models.Index(
                fields=["transaction_subtype", "linked_product", "fee_date", "calculated"],
                name="wbportfolio_transac_1f7a29_idx",
            ),
        ),
    ]



from datetime import timedelta

import pandas as pd
import pytest
from faker import Faker

fake = Faker()


@pytest.mark.django_db
class TestInstrumentMixin:
    def test_get_latest_valid_price(self, product, instrument_price_factory):
        v1 = instrument_price_factory.create(instrument=product)
        v2 = instrument_price_factory.create(instrument=product)
        instrument_price_factory.create(instrument=product, calculated=True)

        assert product.get_latest_valid_price() == v2
        assert product.get_latest_valid_price(v1.date) == v1

    def test_get_earliest_valid_price(self, product, instrument_price_factory):
        instrument_price_factory.create(instrument=product, calculated=True)
        v2 = instrument_price_factory.create(instrument=product)
        v3 = instrument_price_factory.create(instrument=product)

        assert product.get_earliest_valid_price() == v2
        assert product.get_earliest_valid_price(v3.date) == v3

    def test_get_price_range(self, product, instrument_price_factory):
        v_low = instrument_price_factory.create(instrument=product, net_value=1)
        instrument_price_factory.create(instrument=product, net_value=10)
        v_high = instrument_price_factory.create(instrument=product, net_value=100)

        res = product.get_price_range()
        assert res["high"]["price"] == pytest.approx(float(v_high.net_value), rel=1e-4)
        assert res["high"]["date"] == v_high.date
        assert res["low"]["price"] == pytest.approx(float(v_low.net_value), rel=1e-4)
        assert res["low"]["date"] == v_low.date
        assert product.get_price_range(v_low.date - timedelta(days=1)) == dict()

    def test_get_cumulative_shares(self, weekday, product, customer_trade_factory):
        from_date = weekday
        to_date = (weekday + pd.tseries.offsets.BDay(1)).date()
        assert product.get_cumulative_shares(from_date, to_date).to_list() == [0.0, 0.0]

        trade = customer_trade_factory.create(
            transaction_date=weekday, underlying_instrument=product, portfolio=product.primary_portfolio
        )
        assert product.get_cumulative_shares(from_date, to_date).to_list() == [0.0, float(trade.shares)]



from wbcore import filters as wb_filters
from wbfdm.analysis.esg.enums import ESGAggregation

from wbportfolio.models import AssetPosition

from .assets import get_latest_asset_position


class ESGMetricAggregationPortfolioPandasFilterSet(wb_filters.FilterSet):
    date = wb_filters.DateFilter(
        label="Date", lookup_expr="exact", field_name="date", initial=get_latest_asset_position, required=True
    )

    esg_aggregation = wb_filters.ChoiceFilter(
        choices=ESGAggregation.choices(),
        initial=ESGAggregation.GHG_EMISSIONS_SCOPE_1.name,
        required=True,
        method="fake_filter",
    )

    class Meta:
        model = AssetPosition
        fields = {}



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

REGISTER_MENUITEM = MenuItem(
    label="Registers",
    endpoint="wbportfolio:register-list",
    permission=ItemPermission(
        permissions=["wbportfolio.view_register"], method=lambda request: is_internal_user(request.user)
    ),
)



from django.apps import apps as global_apps
from django.db import DEFAULT_DB_ALIAS
from django_celery_beat.models import IntervalSchedule, PeriodicTask

from .constants import DEFAULT_PORTFOLIO_DATA_UPDATE_TASK_INTERVAL_MINUTES


def initialize_task(app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
    PeriodicTask.objects.update_or_create(
        name="Portfolio: Update Company Portfolio Data",
        defaults={
            "name": "Portfolio: Update Company Portfolio Data",
            "interval": IntervalSchedule.objects.get_or_create(
                every=DEFAULT_PORTFOLIO_DATA_UPDATE_TASK_INTERVAL_MINUTES, period=IntervalSchedule.MINUTES
            )[0],
            "task": "wbportfolio.contrib.company_portfolio.tasks.update_all_portfolio_data",
            "crontab": None,
        },
    )
